#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Tom Dwelly
# @Date: 2020-03-03
# @Filename: bhm_spiders_agn.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
# derived from guide.py

# ### flake8: noqa
# isort: skip_file

import peewee
from peewee import JOIN
from peewee import fn

from target_selection.cartons.base import BaseCarton

# general catalogdb imports
from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    EROSITASupersetv1AGN,
)

# imports of existing spectro catalogue
from sdssdb.peewee.sdss5db.catalogdb import (
    CatalogFromSDSS_DR19p_Speclite,
    SDSS_DR19p_Speclite,
)

# additional imports required by bhm_spiders_agn_lsdr10
from sdssdb.peewee.sdss5db.catalogdb import (
    CatalogToLegacy_Survey_DR10,
    Legacy_Survey_DR10,
)

# additional imports required by bhm_spiders_agn_gaiadr3
from sdssdb.peewee.sdss5db.catalogdb import (
    CatalogToGaia_DR3,
    Gaia_DR3,
)

# # additional imports required by bhm_spiders_agn_ps1dr2
# from sdssdb.peewee.sdss5db.catalogdb import (
#     Panstarrs1,
#     CatalogToPanstarrs1,
# )

from target_selection.mag_flux import AB2nMgy
# from target_selection.mag_flux import AB2Jy

# DEBUG STUFF TO USE TEMP TABLE
# CatalogToSDSS_DR19p_Speclite._meta.table_name = 'temp_catalog_to_sdss_dr19p_speclite'
# CatalogToSDSS_DR19p_Speclite._meta._schema = 'sandbox'

# used by cartons that need to compute Galactic latitude:
north_gal_pole_ra = 192.85948  # deg, J2000
north_gal_pole_dec = +27.12825  # deg, J2000

# maskbits to determine if X-ray detection is potentially problematic
# https://wiki.mpe.mpg.de/eRosita/EroFollowup/SDSSV_data_model#Notes_on_ero_flags_column_carrying_eRASS_quality_flagging_information # noqa
ero_flags_mask = (
    0
    + 2**0  # FLAG_SP_SNR - eRASS source is located in/near a supernova remnant # noqa
    + 2**1  # FLAG_SP_BPS - eRASS source is located in/near a bright (X-ray?) point source # noqa
    + 2**2  # FLAG_SP_SCL - eRASS source is located in/near a star cluster # noqa
    + 2**3  # FLAG_SP_LGA - eRASS source is located in/near a local galaxy # noqa
    + 2**4  # FLAG_SP_GC - eRASS source is located in/near a globular cluster # noqa
    + 2
    ** 5  # FLAG_SP_GC_CONS - eRASS source is located in/near a globular cluster (conservative criteria) # noqa
    # + 2**6   # FLAG_NO_RADEC_ERR - eRASS source is missing an estimate of uncertainty on its sky position # noqa
    # + 2**7   # FLAG_NO_EXT_ERR - eRASS source is missing an estimate of uncertainty on its X-ray extent # noqa
    # + 2**8   # FLAG_NO_CTS_ERR - eRASS source is missing an estimate of uncertainty on the number of detected X-ray counts # noqa
    # + 2**9   # FLAG_HARD_SRC   - eRASS source is significantly detected in the 2.3-5keV band (in either eRASS:1 3B or eRASS:3 3B, with DET_LIKE_3 > 8, 30arcsec radius ) # noqa
)

# ############################################
# ############################################
# ############################################
# ############################################
# This provides the following BHM SPIDERS AGN cartons in v1.0:
#  *  bhm_spiders_agn_lsdr10
#  *  bhm_spiders_agn_gaiadr3
#  *  bhm_spiders_agn_sep
#  *  bhm_spiders_agn_tda
#  *  bhm_spiders_agn_hard
#  *  bhm_spiders_agn_lsdr10_d3
#  *  bhm_spiders_agn_gaiadr3_d3
#  *  bhm_spiders_agn_sep_d3
#  *  bhm_spiders_agn_tda_d3
#  *  bhm_spiders_agn_hard_d3
# ############################################
# ############################################
# ############################################
# ############################################

# some reference points for AB->nMgy conversions
# 30.0 AB = 1e-3 nMgy
# 22.5 AB = 1.0 nMgy
# 22.0 AB = 1.58489 nMgy
# 21.5 AB = 2.51189 nMgy
# 21.0 AB = 3.98107 nMgy
# 20.0 AB = 10.0 nMgy
# 18.5 AB = 39.8107 nMgy
# 16.5 AB = 251.189 nMgy
# 14.0 AB = 2511.89 nMgy
# 13.5 AB = 3981.07 nMgy

# some reference points for AB->Jy conversions (for ps1_dr2 fluxes)
# 30.0 AB = 3.631e-9 Jy
# 22.5 AB = 3.631e-6 Jy
# 22.0 AB = 5.754e-6 Jy
# 21.5 AB = 9.120e-6 Jy
# 21.0 AB = 1.445e-5 Jy
# 20.5 AB = 2.291e-5  Jy
# 18.5 AB = 1.445e-4 Jy
# 16.5 AB = 9.120e-4 Jy
# 14.0 AB = 9.120e-3 Jy
# 13.5 AB = 1.445e-2 Jy

# Notes on catalogdb.panstarrs1.flags aka objInfoFlag from ObjectThin
# https://outerspace.stsci.edu/display/PANSTARRS/PS1+ObjectThin+table+fields
# https://outerspace.stsci.edu/display/PANSTARRS/PS1+Object+Flags
# select objects that have the GOOD_STACK flag set:
# Flag name   value      decimal   Notes
# GOOD_STACK  0x08000000 134217728 good-quality object in the stack (> 1 good stack measurement)

# Use these two flags to decide whether to use aper mags or not
# Flag name   value      decimal   Notes
# EXT	      0x00800000   8388608 extended in our data (eg, PS)
# EXT_ALT     0x01000000  16777216 extended in external data (eg, 2MASS)

"""
# noqa
# Notes on how many targets to expect:
 sdss5db=> SELECT ero_version,xmatch_method,xmatch_version,opt_cat,ero_flux_type,count(*)
           FROM erosita_superset_v1_agn GROUP BY ero_version,xmatch_method,xmatch_version,opt_cat,ero_flux_type;
           ero_version           |  xmatch_method  |      xmatch_version      | opt_cat | ero_flux_type |  count
---------------------------------+-----------------+--------------------------+---------+---------------+---------
 eRASS_s1_3B_221031_poscorr      | XPS/NWAY        | JBJWMS_24Nov22           | lsdr10  | 2.3-5keV      |    3433
 eRASS_s3_1B_220829_poscorr_v006 | XPS/NWAY_EROTDA | JWMS_06Oct22_erotda      | lsdr10  | 0.2-2.3keV    |    9703
 eRASS_s3_1B_221007_poscorr_v007 | XPS/NWAY        | JWMS_06Feb23_nomask      | lsdr10  | 0.2-2.3keV    | 1974450
 eRASS_s3_1B_221007_poscorr_v007 | XPS/NWAY        | JWMS_21Oct22             | lsdr10  | 0.2-2.3keV    | 1895479
 eRASS_s3_1B_221007_poscorr_v007 | XPS/NWAY        | JWMS_21Oct22             | lsdr9   | 0.2-2.3keV    |   47172
 eRASS_s3_1B_221007_poscorr_v007 | XPS/NWAY        | JWMS_21Oct22_cw2020_gedr | gedr3   | 0.2-2.3keV    | 1298743
 eRASS_s3_1B_221007_poscorr_v007 | XPS/NWAY        | JWMS_24Oct22             | gedr3   | 0.2-2.3keV    | 2465166
 eRASS_s3_1B_221007_poscorr_v007 | XPS/NWAY        | JWMS_24Oct22_nomask      | lsdr10  | 0.2-2.3keV    | 1937267
 eRASS_s5_V29C                   | XPS/NWAY        | JWTL_Oct22               | gedr3   | 0.2-2.3keV    |    2007
 eRASS_s5_V29C                   | XPS/NWAY        | JWTL_Oct22               | lsdr10  | 0.2-2.3keV    |    5207
(10 rows)
"""

# Notes on avoiding saturated legacysurvey sources
# https://www.legacysurvey.org/dr8/bitmasks/
# Bit   Name       Description
# 0     NPRIMARY   touches a pixel that is outside the BRICK_PRIMARY region of a brick
# 1     BRIGHT     touches a pixel within the locus of a radius-magnitude relation for
#                  Tycho-2 stars or one for Gaia DR2 stars to G < 13
# 2     SATUR_G    touches a pixel that was saturated in at least one g-band image
# 3     SATUR_R    touches a pixel that was saturated in at least one r-band image
# 4     SATUR_Z    touches a pixel that was saturated in at least one z-band image
# 5     ALLMASK_G  touches a pixel that has any of the ALLMASK_G bits set
# 6     ALLMASK_R  touches a pixel that has any of the ALLMASK_R bits set
# 7     ALLMASK_Z  touches a pixel that has any of the ALLMASK_Z bits set
# 8     WISEM1     touches a pixel in a WISEMASK_W1 bright star mask
# 9     WISEM2     touches a pixel in a WISEMASK_W2 bright star mask
# 10    BAILOUT    touches a pixel in a blob where we "bailed out" of source fitting
# 11    MEDIUM     touches a pixel within the locus of a radius-magnitude relation
#                  for Gaia DR2 stars to G < 16
# 12    GALAXY     touches a pixel in an SGA large galaxy
# 13    CLUSTER    touches a pixel in a globular cluster
#
# so, mask to avoid saturated targets is 2**2 + 2**3 + 2**4 = 4+8+16 = 28

#
# END PREAMBLE
# ##################################################################################


class BhmSpidersAgnLsdr10Carton(BaseCarton):
    name = "bhm_spiders_agn_lsdr10"
    category = "science"
    mapper = "BHM"
    program = "bhm_spiders"
    tile = False
    instrument = "BOSS"
    can_offset = True
    only_faintest_cadence = False

    def build_query(self, version_id, query_region=None):
        c = Catalog.alias()
        x = EROSITASupersetv1AGN.alias()
        ls = Legacy_Survey_DR10.alias()
        c2ls = CatalogToLegacy_Survey_DR10.alias()

        instrument = peewee.Value(self.instrument)

        fiberflux_r_max = AB2nMgy(self.parameters["fibermag_r_min"])
        fiberflux_r_min = AB2nMgy(self.parameters["fibermag_r_max"])
        fiberflux_i_max = AB2nMgy(self.parameters["fibermag_i_min"])
        fiberflux_i_min = AB2nMgy(self.parameters["fibermag_i_max"])
        fiberflux_z_max = AB2nMgy(self.parameters["fibermag_z_min"])
        fiberflux_z_min = AB2nMgy(self.parameters["fibermag_z_max"])
        fiberflux_r_max_for_core = AB2nMgy(self.parameters["fibermag_r_min_for_core"])
        fiberflux_r_min_for_core = AB2nMgy(self.parameters["fibermag_r_max_for_core"])
        fiberflux_i_max_for_core = AB2nMgy(self.parameters["fibermag_i_min_for_core"])
        fiberflux_i_min_for_core = AB2nMgy(self.parameters["fibermag_i_max_for_core"])
        fiberflux_z_max_for_core = AB2nMgy(self.parameters["fibermag_z_min_for_core"])
        fiberflux_z_min_for_core = AB2nMgy(self.parameters["fibermag_z_max_for_core"])

        fiberflux_r_min_for_cadence1 = AB2nMgy(self.parameters["fibermag_r_for_cadence1"])
        fiberflux_r_min_for_cadence2 = AB2nMgy(self.parameters["fibermag_r_for_cadence2"])
        fiberflux_i_min_for_cadence1 = AB2nMgy(self.parameters["fibermag_i_for_cadence1"])
        fiberflux_i_min_for_cadence2 = AB2nMgy(self.parameters["fibermag_i_for_cadence2"])
        gaia_g_max_for_cadence1 = self.parameters["gaia_g_max_for_cadence1"]
        gaia_rp_max_for_cadence1 = self.parameters["gaia_rp_max_for_cadence1"]

        # #########################################################################
        # prepare the spectroscopy catalogue
        spec_sn_thresh = self.parameters["spec_sn_thresh"]
        spec_z_err_thresh = self.parameters["spec_z_err_thresh"]

        # SDSS DR19p
        # first downslect only 'good' spectra
        c2s19 = CatalogFromSDSS_DR19p_Speclite.alias()
        ss19 = SDSS_DR19p_Speclite.alias()
        s19 = (
            ss19.select(
                ss19.pk.alias("s19_pk"),
            )
            .where(
                ss19.sn_median_all >= spec_sn_thresh,
                ss19.zwarning == 0,
                ss19.z_err <= spec_z_err_thresh,
                ss19.z_err > 0.0,
                ss19.specprimary > 0,
            )
            .alias("s19")
        )
        #########################################################################

        # compute the abs(Galactic latitude):
        gal_lat = peewee.fn.abs(
            90.0 - peewee.fn.q3c_dist(north_gal_pole_ra, north_gal_pole_dec, c.ra, c.dec)
        )

        # logic is written this backwards way so that a failure to meet any core
        # criterion results in non-core status
        is_core = peewee.Case(
            None,
            (
                (gal_lat < self.parameters["min_gal_lat_for_core"], False),
                (c.dec < self.parameters["min_dec_for_core"], False),
                (x.ero_flux < self.parameters["min_ero_flux_for_core"], False),
                (x.ero_det_like < self.parameters["min_det_like_for_core"], False),
                (
                    ~(
                        (
                            ls.fiberflux_r.between(
                                fiberflux_r_min_for_core, fiberflux_r_max_for_core
                            )
                        )
                        | (
                            ls.fiberflux_i.between(
                                fiberflux_i_min_for_core, fiberflux_i_max_for_core
                            )
                        )
                        | (
                            ls.fiberflux_z.between(
                                fiberflux_z_min_for_core, fiberflux_z_max_for_core
                            )
                        )
                    ),
                    False,
                ),
                (ls.maskbits.bin_and(2**13) > 0, False),  # demote globular clusters and MCs
                (x.ero_flags.bin_and(ero_flags_mask) > 0, False),  # demote problematic X-ray data
            ),
            True,
        )

        # value = peewee.Value(self.parameters.get('value', 1.0)).cast('float')
        value = peewee.Case(None, ((is_core, self.parameters.get("value", 1.0)),), 0.0).cast(
            "float"
        )

        # priority is determined from individual target properties
        # start with a priority floor value (per carton)
        # then increment if any conditions are met:
        # add +dpriority_match_flags if target is a secondary cross-match (match_flag > 1)
        # add +dpriority_det_like if target has a low value of ero_det_like
        # add +dpriority_has_spec if target has existing good SDSS spectroscopy
        priority_1 = peewee.Case(None, ((x.xmatch_flags > 1, 1),), 0)
        priority_2 = peewee.Case(
            None, ((x.ero_det_like < self.parameters["det_like_for_priority"], 1),), 0
        )
        priority_3 = peewee.Case(None, ((s19.c.s19_pk.is_null(False), 1),), 0)
        priority_4 = peewee.Case(None, ((is_core, 0),), 1)
        priority_5 = peewee.Case(None, ((x.ero_flags.bin_and(2**9) == 0, 1),), 0)

        # priority = fn.max(
        priority = (
            self.parameters["priority_floor"]
            + priority_1 * self.parameters["dpriority_match_flags"]
            + priority_2 * self.parameters["dpriority_det_like"]
            + priority_3 * self.parameters["dpriority_has_spec"]
            + priority_4 * self.parameters["dpriority_non_core"]
            + priority_5 * self.parameters["dpriority_not_hard"]
        )

        # choose cadence based on fiber magnitude in r-band
        cadence1 = self.parameters["cadence1"]
        cadence2 = self.parameters["cadence2"]
        cadence3 = self.parameters["cadence3"]
        # cadence4 = 'unknown_cadence'
        cadence = peewee.Case(
            None,
            (
                (
                    (
                        (ls.fiberflux_r > fiberflux_r_min_for_cadence1)
                        | (ls.fiberflux_i > fiberflux_i_min_for_cadence1)
                        | (ls.gaia_phot_g_mean_mag.between(0.1, gaia_g_max_for_cadence1))
                        | (ls.gaia_phot_rp_mean_mag.between(0.1, gaia_rp_max_for_cadence1))
                    ),
                    cadence1,
                ),
                (
                    (ls.fiberflux_r > fiberflux_r_min_for_cadence2)
                    | (ls.fiberflux_i > fiberflux_i_min_for_cadence2),
                    cadence2,
                ),
                #  ((ls.fiberflux_r < fiberflux_r_min_for_cadence2) &
                #   (ls.fiberflux_i < fiberflux_i_min_for_cadence2),
                #    cadence3),
            ),
            cadence3,
        )
        #    cadence4)

        # compute transformed SDSS mags for pointlike and extended sources separately
        # transform the legacysurvey grz into sdss psfmag griz

        # extract coeffs from fit logs via:
        # awk 'BEGIN {print("coeffs = {")} /POLYFIT/{ if($3~/sdss_psfmag/){pe="p"} else if ($3~/sdss_fiber2mag/){pe="e"} else{pe="error"}; printf("\"%s%d_%s\": %s,\n", substr($3,length($3)), $8, pe, $10)} END {print("}")}'  bhm_spiders_agn_lsdr8_*/lsdr8_*mag_to_sdss_*mag_?_results.log  # noqa
        coeffs = {
            "g2_e": -0.113816,
            "g1_e": 0.317176,
            "g0_e": 0.094145,
            "i2_e": -0.415858,
            "i1_e": 0.168922,
            "i0_e": -0.010771,
            "r2_e": 0.029398,
            "r1_e": -0.019938,
            "r0_e": 0.354042,
            "z2_e": -0.111262,
            "z1_e": 0.237656,
            "z0_e": 0.148923,
            "g2_p": 0.187193,
            "g1_p": -0.184362,
            "g0_p": 0.049492,
            "i2_p": -0.098979,
            "i1_p": -0.405518,
            "i0_p": 0.009688,
            "r2_p": -0.001935,
            "r1_p": 0.098201,
            "r0_p": 0.050321,
            "z2_p": -0.034163,
            "z1_p": 0.109878,
            "z0_p": -0.030167,
        }

        nMgy_min = 1e-3  # equiv to AB=30
        # pointlike - start from ls8 (psf)fluxes
        g0_p = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_g))
        r0_p = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_r))
        i0_p = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_i))
        z0_p = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_z))
        g_r_p = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_g) / peewee.fn.greatest(nMgy_min, ls.flux_r)
        )
        r_z_p = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_r) / peewee.fn.greatest(nMgy_min, ls.flux_z)
        )

        # extended - start from ls8 fiberfluxes
        g0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_g))
        r0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_r))
        i0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_i))
        z0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_z))
        g_r_e = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.fiberflux_g)
            / peewee.fn.greatest(nMgy_min, ls.fiberflux_r)
        )
        r_z_e = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.fiberflux_r)
            / peewee.fn.greatest(nMgy_min, ls.fiberflux_z)
        )

        g_p = g0_p + coeffs["g0_p"] + coeffs["g1_p"] * g_r_p + coeffs["g2_p"] * g_r_p * g_r_p
        r_p = r0_p + coeffs["r0_p"] + coeffs["r1_p"] * g_r_p + coeffs["r2_p"] * g_r_p * g_r_p
        i_p = r0_p + coeffs["i0_p"] + coeffs["i1_p"] * r_z_p + coeffs["i2_p"] * r_z_p * r_z_p
        z_p = z0_p + coeffs["z0_p"] + coeffs["z1_p"] * r_z_p + coeffs["z2_p"] * r_z_p * r_z_p

        g_e = g0_e + coeffs["g0_e"] + coeffs["g1_e"] * g_r_e + coeffs["g2_e"] * g_r_e * g_r_e
        r_e = r0_e + coeffs["r0_e"] + coeffs["r1_e"] * g_r_e + coeffs["r2_e"] * g_r_e * g_r_e
        i_e = r0_e + coeffs["i0_e"] + coeffs["i1_e"] * r_z_e + coeffs["i2_e"] * r_z_e * r_z_e
        z_e = z0_e + coeffs["z0_e"] + coeffs["z1_e"] * r_z_e + coeffs["z2_e"] * r_z_e * r_z_e

        # validity checks - set limits semi-manually
        g_r_p_min = -0.25
        g_r_p_max = 1.75
        r_z_p_min = -0.5
        r_z_p_max = 2.5
        g_r_e_min = 0.0
        g_r_e_max = 1.75
        r_z_e_min = 0.2
        r_z_e_max = 1.6
        valid_p = (
            g0_p.between(0.1, 29.9)
            & r0_p.between(0.1, 29.9)
            & z0_p.between(0.1, 29.9)
            & g_r_p.between(g_r_p_min, g_r_p_max)
            & r_z_p.between(r_z_p_min, r_z_p_max)
        )
        valid_e = (
            g0_e.between(0.1, 29.9)
            & r0_e.between(0.1, 29.9)
            & z0_e.between(0.1, 29.9)
            & g_r_e.between(g_r_e_min, g_r_e_max)
            & r_z_e.between(r_z_e_min, r_z_e_max)
        )

        # We want to switch between psfmags and fibermags depending on
        # ls.type parameter (PSF or extended)
        # For 'PSF' targets, we use psfmags, but for extended sources use fiber2mags
        opt_prov = peewee.Case(
            None,
            (
                ((ls.type == "PSF") & valid_p, "sdss_psfmag_from_lsdr10"),
                ((ls.type != "PSF") & valid_e, "sdss_fiber2mag_from_lsdr10"),
            ),
            "undefined",
        )

        magnitude_g = peewee.Case(
            None,
            (
                ((ls.type == "PSF") & valid_p, g_p.cast("float")),
                ((ls.type != "PSF") & valid_e, g_e.cast("float")),
            ),
            "NaN",
        )
        magnitude_r = peewee.Case(
            None,
            (
                ((ls.type == "PSF") & valid_p, r_p.cast("float")),
                ((ls.type != "PSF") & valid_e, r_e.cast("float")),
            ),
            "NaN",
        )
        magnitude_i = peewee.Case(
            None,
            (
                ((ls.type == "PSF") & valid_p, i_p.cast("float")),
                ((ls.type != "PSF") & valid_e, i_e.cast("float")),
            ),
            "NaN",
        )
        magnitude_z = peewee.Case(
            None,
            (
                ((ls.type == "PSF") & valid_p, z_p.cast("float")),
                ((ls.type != "PSF") & valid_e, z_e.cast("float")),
            ),
            "NaN",
        )

        magnitude_gaia_g = peewee.Case(
            None, ((ls.gaia_phot_g_mean_mag.between(0.1, 29.9), ls.gaia_phot_g_mean_mag),), "NaN"
        )
        magnitude_gaia_bp = peewee.Case(
            None, ((ls.gaia_phot_bp_mean_mag.between(0.1, 29.9), ls.gaia_phot_bp_mean_mag),), "NaN"
        )
        magnitude_gaia_rp = peewee.Case(
            None, ((ls.gaia_phot_rp_mean_mag.between(0.1, 29.9), ls.gaia_phot_rp_mean_mag),), "NaN"
        )

        query = (
            c.select(
                c.catalogid.alias("catalogid"),
                priority.alias("priority"),
                value.alias("value"),
                cadence.alias("cadence"),
                instrument.alias("instrument"),
                opt_prov.alias("optical_prov"),
                magnitude_g.alias("g"),
                magnitude_r.alias("r"),
                magnitude_i.alias("i"),
                magnitude_z.alias("z"),
                magnitude_gaia_g.alias("gaia_g"),
                magnitude_gaia_bp.alias("bp"),
                magnitude_gaia_rp.alias("rp"),
                ls.ls_id.alias("ls_id"),  # extra
                ls.gaia_dr3_source_id.alias("gaia_dr3_source_id"),  # extra
                ls.gaia_dr2_source_id.alias("gaia_dr2_source_id"),  # extra
                x.ero_detuid.alias("ero_detuid"),  # extra
                x.ero_flux.alias("ero_flux"),  # extra
                x.ero_det_like.alias("ero_det_like"),  # extra
                x.ero_flags.alias("ero_flags"),  # extra
                x.xmatch_flags.alias("xmatch_flags"),  # extra
                x.xmatch_metric.alias("xmatch_metric"),  # extra
                s19.c.s19_pk.alias("sdss_dr19p_speclite_pk"),  # extra
                c.ra.alias("ra"),  # extra
                c.dec.alias("dec"),  # extra
                g0_p.alias("ls10_mag_g"),  # extra
                r0_p.alias("ls10_mag_r"),  # extra
                i0_p.alias("ls10_mag_i"),  # extra
                z0_p.alias("ls10_mag_z"),  # extra
                g0_e.alias("ls10_fibermag_g"),  # extra
                r0_e.alias("ls10_fibermag_r"),  # extra
                i0_e.alias("ls10_fibermag_i"),  # extra
                z0_e.alias("ls10_fibermag_z"),  # extra
                ls.nobs_g.alias("ls10_nobs_g"),
                ls.nobs_r.alias("ls10_nobs_r"),
                ls.nobs_i.alias("ls10_nobs_i"),
                ls.nobs_z.alias("ls10_nobs_z"),
                ls.type.alias("ls10_type"),  # extra
                ls.shape_r.alias("ls10_shape_r"),  # extra
                is_core.alias("is_core"),  # extra
                ls.ref_cat.alias("ls_ref_cat"),  # extra
                ls.ref_id.alias("ls_ref_id"),  # extra
                ls.maskbits.alias("ls_maskbits"),  # extra
                ls.fitbits.alias("ls_fitbits"),  # extra
                gal_lat.alias("abs_gal_lat"),  # extra
            )
            .join(c2ls)
            .join(ls)
            .join(x, on=(ls.ls_id == x.ls_id))
            # start joining the spectroscopy
            .switch(c)
            .join(c2s19, JOIN.LEFT_OUTER)
            .join(s19, JOIN.LEFT_OUTER, on=(s19.c.s19_pk == c2s19.target_id))
            # finished joining the spectroscopy
            # admin criteria
            .where(
                c.version_id == version_id,
                c2ls.version_id == version_id,
                fn.coalesce(c2s19.version_id, version_id) == version_id,
                c2ls.best >> True,
                # fn.coalesce(c2s19.best, True) >> True,
            )
            # science criteria
            .where(
                (x.ero_version == self.parameters["ero_version"]),
                (x.xmatch_method == self.parameters["xmatch_method"]),
                (
                    (x.xmatch_version == self.parameters["xmatch_version1"])
                    | (x.xmatch_version == self.parameters["xmatch_version2"])
                ),
                (
                    (x.opt_cat == self.parameters["opt_cat1"])
                    | (x.opt_cat == self.parameters["opt_cat2"])
                ),
                (x.xmatch_metric >= self.parameters["p_any_min"]),
                (
                    (ls.fiberflux_r.between(fiberflux_r_min, fiberflux_r_max))
                    | (ls.fiberflux_i.between(fiberflux_i_min, fiberflux_i_max))
                    | (ls.fiberflux_z.between(fiberflux_z_min, fiberflux_z_max))
                ),
                (x.ero_det_like > self.parameters["det_like_min"]),
                # (ls.maskbits.bin_and(2**2 + 2**3 + 2**4) == 0),  # avoid saturated sources
                # avoid bright stars and globular clusters:
                # (ls.maskbits.bin_and(2**1 + 2**13) == 0),
                # always avoid very bright stars, see https://www.legacysurvey.org/dr10/bitmasks/
                (ls.maskbits.bin_and(2**1) == 0),
                # (ls.nobs_r > 0),                        # always require r-band coverage
                # ((ls.nobs_g > 0) | (ls.nobs_z > 0)),    # plus at least one other optical band
                # gaia safety checks to avoid bad ls photometry
                ~(ls.gaia_phot_g_mean_mag.between(0.1, self.parameters["gaia_g_mag_limit"])),
                ~(ls.gaia_phot_rp_mean_mag.between(0.1, self.parameters["gaia_rp_mag_limit"])),
            )
            # .group_by(ls)   # avoid duplicates - we trust the legacy survey entries
            .distinct([ls.ls_id])  # avoid duplicates - we trust the legacy survey entries
        )

        if self.only_faintest_cadence:
            query = query.where(cadence == cadence3)

        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    c.ra, c.dec, query_region[0], query_region[1], query_region[2]
                )
            )

        return query


#
# END BhmSpidersAgnLsdr10Carton
# ##################################################################################
class BhmSpidersAgnLsdr10D3Carton(BhmSpidersAgnLsdr10Carton):
    name = "bhm_spiders_agn_lsdr10_d3"
    only_faintest_cadence = True


# we can get away with just inheriting the selection code from
# the lsdr10 hemisphere match and adjusting the parameters only
# ##################################################################################
class BhmSpidersAgnHardCarton(BhmSpidersAgnLsdr10Carton):
    name = "bhm_spiders_agn_hard"


class BhmSpidersAgnHardD3Carton(BhmSpidersAgnLsdr10Carton):
    name = "bhm_spiders_agn_hard_d3"
    only_faintest_cadence = True


# # Testing of the North part of lsdr10 (i.e. dr9)
# # we can get away with just inheriting the selection code from
# # the lsdr10 hemisphere match and adjusting the parameters only
# # ##################################################################################
# class BhmSpidersAgnLsdr10NorthCarton(BhmSpidersAgnLsdr10Carton):
#     name = 'bhm_spiders_agn_lsdr10_north'


class BhmSpidersAgnGaiadr3Carton(BaseCarton):
    name = "bhm_spiders_agn_gaiadr3"
    category = "science"
    mapper = "BHM"
    program = "bhm_spiders"
    tile = False
    instrument = "BOSS"
    can_offset = True
    only_faintest_cadence = False

    def build_query(self, version_id, query_region=None):
        c = Catalog.alias()
        x = EROSITASupersetv1AGN.alias()
        g3 = Gaia_DR3.alias()
        c2g3 = CatalogToGaia_DR3.alias()

        instrument = peewee.Value(self.instrument)

        gaia_g_max_for_cadence1 = self.parameters["gaia_g_max_for_cadence1"]
        gaia_rp_max_for_cadence1 = self.parameters["gaia_rp_max_for_cadence1"]
        gaia_g_max_for_cadence2 = self.parameters["gaia_g_max_for_cadence2"]
        gaia_rp_max_for_cadence2 = self.parameters["gaia_rp_max_for_cadence2"]

        # these control matching to spectroscopy
        spec_sn_thresh = self.parameters["spec_sn_thresh"]
        spec_z_err_thresh = self.parameters["spec_z_err_thresh"]

        # #########################################################################
        # prepare the spectroscopy catalogues

        # SDSS DR19p
        # downslect only 'good' spectra
        c2s19 = CatalogFromSDSS_DR19p_Speclite.alias()
        ss19 = SDSS_DR19p_Speclite.alias()
        s19 = (
            ss19.select(
                ss19.pk.alias("s19_pk"),
            )
            .where(
                ss19.sn_median_all >= spec_sn_thresh,
                ss19.zwarning == 0,
                ss19.z_err <= spec_z_err_thresh,
                ss19.z_err > 0.0,
                ss19.specprimary > 0,
            )
            .alias("s19")
        )
        # #########################################################################

        # compute the abs(Galactic latitude):
        gal_lat = peewee.fn.abs(
            90.0 - peewee.fn.q3c_dist(north_gal_pole_ra, north_gal_pole_dec, c.ra, c.dec)
        )

        is_core = peewee.Case(
            None,
            (
                (gal_lat < self.parameters["min_gal_lat_for_core"], False),
                (c.dec < self.parameters["min_dec_for_core"], False),
                (x.ero_flux < self.parameters["min_ero_flux_for_core"], False),
                (x.ero_det_like < self.parameters["min_det_like_for_core"], False),
                (g3.phot_g_mean_mag < self.parameters["min_gaia_g_for_core"], False),
                (g3.phot_g_mean_mag > self.parameters["max_gaia_g_for_core"], False),
            ),
            True,
        )

        # value = peewee.Value(self.parameters.get('value', 1.0)).cast('float')
        value = peewee.Case(
            None,
            #  ((gal_lat > self.parameters['in_plane_lat_cut'],
            #    self.parameters.get('value', 1.0)),),
            ((is_core, self.parameters.get("value", 1.0)),),
            0.0,
        ).cast("float")

        # priority is determined by target properties
        # start with a priority floor value (per carton)
        # then increment if any conditions are met:
        # add +dpriority_match_flags if target is a secondary cross-match (match_flag > 1)
        # add +dpriority_det_like if target has a low value of ero_det_like
        # add +dpriority_has_spec if target has existing good SDSS spectroscopy

        priority_1 = peewee.Case(None, ((x.xmatch_flags > 1, 1),), 0)
        priority_2 = peewee.Case(
            None, ((x.ero_det_like < self.parameters["det_like_for_priority"], 1),), 0
        )
        priority_3 = peewee.Case(None, ((s19.c.s19_pk.is_null(False), 1),), 0)
        priority_4 = peewee.Case(None, ((is_core, 0),), 1)
        priority_5 = peewee.Case(None, ((x.ero_flags.bin_and(2**9) == 0, 1),), 0)

        priority = (
            self.parameters["priority_floor"]
            + priority_1 * self.parameters["dpriority_match_flags"]
            + priority_2 * self.parameters["dpriority_det_like"]
            + priority_3 * self.parameters["dpriority_has_spec"]
            + priority_4 * self.parameters["dpriority_non_core"]
            + priority_5 * self.parameters["dpriority_not_hard"]
        )

        # choose cadence based on magnitude in Gaia G and RP-bands
        cadence1 = self.parameters["cadence1"]
        cadence2 = self.parameters["cadence2"]
        cadence3 = self.parameters["cadence3"]
        cadence4 = "unknown_cadence"
        cadence = peewee.Case(
            None,
            (
                (
                    (g3.phot_g_mean_mag < gaia_g_max_for_cadence1)
                    | (g3.phot_rp_mean_mag < gaia_rp_max_for_cadence1),
                    cadence1,
                ),
                (
                    (g3.phot_g_mean_mag < gaia_g_max_for_cadence2)
                    | (g3.phot_rp_mean_mag < gaia_rp_max_for_cadence2),
                    cadence2,
                ),
                (
                    (g3.phot_g_mean_mag >= gaia_g_max_for_cadence2)
                    & (g3.phot_rp_mean_mag >= gaia_rp_max_for_cadence2),
                    cadence3,
                ),
            ),
            cadence4,
        )

        # compute transformed SDSS mags
        # transform the Gaia dr3 G,BP,RP into sdss psfmag griz
        # piecewise transformation either side of BP-RP=1.8
        # fit to blue end is cubic, fit to red end is quadratic
        # awk 'BEGIN {print("coeffs = {")} /POLYFIT/{ if(FILENAME~/_red/){pe="red"} else if (FILENAME~/_blue/){pe="blue"} else{pe="error"}; printf("\"%s%d_%s\": %s,\n", substr($3,length($3)), $8, pe, $10)} END {print("}")}'  bhm_spiders_agn_gaiadr2_red/gdr2_*mag_to_sdss_*mag_?_results.log bhm_spiders_agn_gaiadr2_blue/gdr2_*mag_to_sdss_*mag_?_results.log  # noqa
        coeffs = {
            "g2_red": 0.081178,
            "g1_red": 0.355677,
            "g0_red": 0.510306,
            "i2_red": 0.048864,
            "i1_red": -0.287475,
            "i0_red": -0.336712,
            "r2_red": 0.028080,
            "r1_red": 0.542331,
            "r0_red": -1.055168,
            "z2_red": -0.131385,
            "z1_red": 0.302555,
            "z0_red": -1.381648,
            "g3_blue": 0.639054,
            "g2_blue": -1.739187,
            "g1_blue": 1.420330,
            "g0_blue": -0.194071,
            "i3_blue": 0.780585,
            "i2_blue": -2.549848,
            "i1_blue": 1.489880,
            "i0_blue": -0.241381,
            "r3_blue": 0.575494,
            "r2_blue": -2.077000,
            "r1_blue": 1.573302,
            "r0_blue": -0.295026,
            "z3_blue": 1.064986,
            "z2_blue": -3.162969,
            "z1_blue": 1.493750,
            "z0_blue": -0.199582,
        }

        g_blue = (
            g3.phot_g_mean_mag
            + coeffs["g0_blue"]
            + coeffs["g1_blue"] * g3.bp_rp
            + coeffs["g2_blue"] * g3.bp_rp * g3.bp_rp
            + coeffs["g3_blue"] * g3.bp_rp * g3.bp_rp * g3.bp_rp
        )
        r_blue = (
            g3.phot_g_mean_mag
            + coeffs["r0_blue"]
            + coeffs["r1_blue"] * g3.bp_rp
            + coeffs["r2_blue"] * g3.bp_rp * g3.bp_rp
            + coeffs["r3_blue"] * g3.bp_rp * g3.bp_rp * g3.bp_rp
        )
        i_blue = (
            g3.phot_g_mean_mag
            + coeffs["i0_blue"]
            + coeffs["i1_blue"] * g3.bp_rp
            + coeffs["i2_blue"] * g3.bp_rp * g3.bp_rp
            + coeffs["i3_blue"] * g3.bp_rp * g3.bp_rp * g3.bp_rp
        )
        z_blue = (
            g3.phot_g_mean_mag
            + coeffs["z0_blue"]
            + coeffs["z1_blue"] * g3.bp_rp
            + coeffs["z2_blue"] * g3.bp_rp * g3.bp_rp
            + coeffs["z3_blue"] * g3.bp_rp * g3.bp_rp * g3.bp_rp
        )

        g_red = (
            g3.phot_g_mean_mag
            + coeffs["g0_red"]
            + coeffs["g1_red"] * g3.bp_rp
            + coeffs["g2_red"] * g3.bp_rp * g3.bp_rp
        )
        r_red = (
            g3.phot_g_mean_mag
            + coeffs["r0_red"]
            + coeffs["r1_red"] * g3.bp_rp
            + coeffs["r2_red"] * g3.bp_rp * g3.bp_rp
        )
        i_red = (
            g3.phot_g_mean_mag
            + coeffs["i0_red"]
            + coeffs["i1_red"] * g3.bp_rp
            + coeffs["i2_red"] * g3.bp_rp * g3.bp_rp
        )
        z_red = (
            g3.phot_g_mean_mag
            + coeffs["z0_red"]
            + coeffs["z1_red"] * g3.bp_rp
            + coeffs["z2_red"] * g3.bp_rp * g3.bp_rp
        )

        # validity checks - set limits semi-manually
        bp_rp_min = 0.0
        bp_rp_max = 3.0
        valid = (
            g3.phot_g_mean_mag.between(0.1, 29.9)
            & g3.phot_bp_mean_mag.between(0.1, 29.9)
            & g3.phot_rp_mean_mag.between(0.1, 29.9)
            & g3.bp_rp.between(bp_rp_min, bp_rp_max)
        )
        opt_prov = peewee.Case(None, ((valid, "sdss_psfmag_from_gdr3"),), "undefined")
        magnitude_g = peewee.Case(
            None,
            (
                (valid & (g3.bp_rp < 1.8), g_blue),
                (valid & (g3.bp_rp > 1.8), g_red),
            ),
            "NaN",
        )
        magnitude_r = peewee.Case(
            None,
            (
                (valid & (g3.bp_rp < 1.8), r_blue),
                (valid & (g3.bp_rp > 1.8), r_red),
            ),
            "NaN",
        )
        magnitude_i = peewee.Case(
            None,
            (
                (valid & (g3.bp_rp < 1.8), i_blue),
                (valid & (g3.bp_rp > 1.8), i_red),
            ),
            "NaN",
        )
        magnitude_z = peewee.Case(
            None,
            (
                (valid & (g3.bp_rp < 1.8), z_blue),
                (valid & (g3.bp_rp > 1.8), z_red),
            ),
            "NaN",
        )

        query = (
            c.select(
                c.catalogid.alias("catalogid"),
                x.ero_detuid.alias("ero_detuid"),  # extra
                x.ero_flux.alias("ero_flux"),  # extra
                x.ero_det_like.alias("ero_det_like"),  # extra
                g3.source_id.alias("gaia_dr3_source_id"),  # extra
                s19.c.s19_pk.alias("sdss_dr19p_speclite_pk"),  # extra
                c.ra.alias("ra"),  # extra
                c.dec.alias("dec"),  # extra
                priority.alias("priority"),
                value.alias("value"),
                cadence.alias("cadence"),
                instrument.alias("instrument"),
                opt_prov.alias("optical_prov"),
                magnitude_g.alias("g"),
                magnitude_r.alias("r"),
                magnitude_i.alias("i"),
                magnitude_z.alias("z"),
                g3.phot_g_mean_mag.alias("gaia_g"),
                g3.phot_bp_mean_mag.alias("bp"),
                g3.phot_rp_mean_mag.alias("rp"),
                is_core.alias("is_core"),  # extra
                gal_lat.alias("abs_gal_lat"),  # extra
                x.xmatch_version.alias("xmatch_version"),  # extra
            )
            .join(c2g3)
            .where(
                c.version_id == version_id,
                c2g3.version_id == version_id,
                fn.coalesce(c2s19.version_id, version_id) == version_id,
                c2g3.best >> True,
            )
            .join(g3)
            .join(x, on=(g3.source_id == x.gaia_dr3_source_id))
            # start joining the spectroscopy
            .switch(c)
            .join(c2s19, JOIN.LEFT_OUTER)
            .join(s19, JOIN.LEFT_OUTER, on=(s19.c.s19_pk == c2s19.target_id))
            # finished joining the spectroscopy
            .where(
                (x.ero_version == self.parameters["ero_version"]),
                (
                    (x.xmatch_method == self.parameters["xmatch_method1"])
                    | (x.xmatch_method == self.parameters["xmatch_method2"])
                ),
                (
                    (x.xmatch_version == self.parameters["xmatch_version1"])
                    | (x.xmatch_version == self.parameters["xmatch_version2"])
                ),
                (x.opt_cat == self.parameters["opt_cat"]),
                (x.xmatch_metric >= self.parameters["p_any_min"]),
                (g3.phot_g_mean_mag > self.parameters["gaia_g_mag_limit"]),
                (g3.phot_rp_mean_mag > self.parameters["gaia_rp_mag_limit"]),
                (x.ero_det_like > self.parameters["det_like_min"]),
            )
            .distinct([g3.source_id])  # avoid duplicates - we trust the gaia ids
        )

        if self.only_faintest_cadence:
            query = query.where(cadence == cadence3)

        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    c.ra, c.dec, query_region[0], query_region[1], query_region[2]
                )
            )

        return query


#
# END BhmSpidersAgnGaiadr3Carton


# For the version that goes via catwise202,
# we can get away with just inheriting the selection code from
# the gaia dr3 hemisphere match and adjusting the parameters only
# ##################################################################################
class BhmSpidersAgnGaiadr3viaCW2020Carton(BhmSpidersAgnGaiadr3Carton):
    name = "bhm_spiders_agn_gaiadr3_viacw2020"


# ##################################################################################
class BhmSpidersAgnGaiadr3BothCarton(BhmSpidersAgnGaiadr3Carton):
    name = "bhm_spiders_agn_gaiadr3_both"


# ##################################################################################
# we can get away with just inheriting the selection code from
# the gaia dr3 hemisphere match and adjusting the parameters only
class BhmSpidersAgnSepCarton(BhmSpidersAgnGaiadr3Carton):
    name = "bhm_spiders_agn_sep"


# ##################################################################################
class BhmSpidersAgnGaiadr3D3CartonCarton(BhmSpidersAgnGaiadr3Carton):
    name = "bhm_spiders_agn_gaiadr3_d3"
    only_faintest_cadence = True


# ##################################################################################
class BhmSpidersAgnSepD3Carton(BhmSpidersAgnGaiadr3Carton):
    name = "bhm_spiders_agn_sep_d3"
    only_faintest_cadence = True


class BhmSpidersAgnTdaCarton(BaseCarton):
    name = "bhm_spiders_agn_tda"
    category = "science"
    mapper = "BHM"
    program = "bhm_spiders"
    tile = False
    instrument = "BOSS"
    can_offset = True
    only_faintest_cadence = False

    def build_query(self, version_id, query_region=None):
        c = Catalog.alias()
        x = EROSITASupersetv1AGN.alias()
        ls = Legacy_Survey_DR10.alias()
        c2ls = CatalogToLegacy_Survey_DR10.alias()

        instrument = peewee.Value(self.instrument)

        fiberflux_r_max = AB2nMgy(self.parameters["fibermag_r_min"])
        fiberflux_r_min = AB2nMgy(self.parameters["fibermag_r_max"])
        fiberflux_i_max = AB2nMgy(self.parameters["fibermag_i_min"])
        fiberflux_i_min = AB2nMgy(self.parameters["fibermag_i_max"])
        fiberflux_z_max = AB2nMgy(self.parameters["fibermag_z_min"])
        fiberflux_z_min = AB2nMgy(self.parameters["fibermag_z_max"])
        fiberflux_r_min_for_cadence1 = AB2nMgy(self.parameters["fibermag_r_for_cadence1"])
        fiberflux_r_min_for_cadence2 = AB2nMgy(self.parameters["fibermag_r_for_cadence2"])
        gaia_g_max_for_cadence1 = self.parameters["gaia_g_max_for_cadence1"]
        gaia_rp_max_for_cadence1 = self.parameters["gaia_rp_max_for_cadence1"]

        # choose cadence based on fiber magnitude in r-band + gaia G,RP
        cadence1 = self.parameters["cadence1"]
        cadence2 = self.parameters["cadence2"]
        cadence3 = self.parameters["cadence3"]
        cadence4 = "unknown_cadence"
        cadence = peewee.Case(
            None,
            (
                (
                    (
                        (ls.fiberflux_r > fiberflux_r_min_for_cadence1)
                        | (ls.gaia_phot_g_mean_mag.between(0.1, gaia_g_max_for_cadence1))
                        | (ls.gaia_phot_rp_mean_mag.between(0.1, gaia_rp_max_for_cadence1))
                    ),
                    cadence1,
                ),
                (ls.fiberflux_r > fiberflux_r_min_for_cadence2, cadence2),
                (ls.fiberflux_r <= fiberflux_r_min_for_cadence2, cadence3),
            ),
            cadence4,
        )

        value = peewee.Value(self.parameters["value"]).cast("float")
        priority = peewee.Value(self.parameters["priority_floor"]).cast("integer")

        # compute transformed SDSS mags for pointlike and extended sources separately
        # transform the legacysurvey grz into sdss psfmag griz

        # extract coeffs from fit logs via:
        # awk 'BEGIN {print("coeffs = {")} /POLYFIT/{ if($3~/sdss_psfmag/){pe="p"} else if ($3~/sdss_fiber2mag/){pe="e"} else{pe="error"}; printf("\"%s%d_%s\": %s,\n", substr($3,length($3)), $8, pe, $10)} END {print("}")}'  bhm_spiders_agn_lsdr8_*/lsdr8_*mag_to_sdss_*mag_?_results.log  # noqa
        coeffs = {
            "g2_e": -0.113816,
            "g1_e": 0.317176,
            "g0_e": 0.094145,
            "i2_e": -0.415858,
            "i1_e": 0.168922,
            "i0_e": -0.010771,
            "r2_e": 0.029398,
            "r1_e": -0.019938,
            "r0_e": 0.354042,
            "z2_e": -0.111262,
            "z1_e": 0.237656,
            "z0_e": 0.148923,
            "g2_p": 0.187193,
            "g1_p": -0.184362,
            "g0_p": 0.049492,
            "i2_p": -0.098979,
            "i1_p": -0.405518,
            "i0_p": 0.009688,
            "r2_p": -0.001935,
            "r1_p": 0.098201,
            "r0_p": 0.050321,
            "z2_p": -0.034163,
            "z1_p": 0.109878,
            "z0_p": -0.030167,
        }

        nMgy_min = 1e-3  # equiv to AB=30
        # pointlike - start from ls8 (psf)fluxes
        g0_p = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_g))
        r0_p = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_r))
        i0_p = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_i))
        z0_p = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_z))
        g_r_p = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_g) / peewee.fn.greatest(nMgy_min, ls.flux_r)
        )
        r_z_p = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_r) / peewee.fn.greatest(nMgy_min, ls.flux_z)
        )

        # extended - start from ls8 fiberfluxes
        g0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_g))
        r0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_r))
        i0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_i))
        z0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_z))
        g_r_e = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.fiberflux_g)
            / peewee.fn.greatest(nMgy_min, ls.fiberflux_r)
        )
        r_z_e = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.fiberflux_r)
            / peewee.fn.greatest(nMgy_min, ls.fiberflux_z)
        )

        g_p = g0_p + coeffs["g0_p"] + coeffs["g1_p"] * g_r_p + coeffs["g2_p"] * g_r_p * g_r_p
        r_p = r0_p + coeffs["r0_p"] + coeffs["r1_p"] * g_r_p + coeffs["r2_p"] * g_r_p * g_r_p
        i_p = r0_p + coeffs["i0_p"] + coeffs["i1_p"] * r_z_p + coeffs["i2_p"] * r_z_p * r_z_p
        z_p = z0_p + coeffs["z0_p"] + coeffs["z1_p"] * r_z_p + coeffs["z2_p"] * r_z_p * r_z_p

        g_e = g0_e + coeffs["g0_e"] + coeffs["g1_e"] * g_r_e + coeffs["g2_e"] * g_r_e * g_r_e
        r_e = r0_e + coeffs["r0_e"] + coeffs["r1_e"] * g_r_e + coeffs["r2_e"] * g_r_e * g_r_e
        i_e = r0_e + coeffs["i0_e"] + coeffs["i1_e"] * r_z_e + coeffs["i2_e"] * r_z_e * r_z_e
        z_e = z0_e + coeffs["z0_e"] + coeffs["z1_e"] * r_z_e + coeffs["z2_e"] * r_z_e * r_z_e

        # validity checks - set limits semi-manually
        g_r_p_min = -0.25
        g_r_p_max = 1.75
        r_z_p_min = -0.5
        r_z_p_max = 2.5
        g_r_e_min = 0.0
        g_r_e_max = 1.75
        r_z_e_min = 0.2
        r_z_e_max = 1.6
        valid_p = (
            g0_p.between(0.1, 29.9)
            & r0_p.between(0.1, 29.9)
            & z0_p.between(0.1, 29.9)
            & g_r_p.between(g_r_p_min, g_r_p_max)
            & r_z_p.between(r_z_p_min, r_z_p_max)
        )
        valid_e = (
            g0_e.between(0.1, 29.9)
            & r0_e.between(0.1, 29.9)
            & z0_e.between(0.1, 29.9)
            & g_r_e.between(g_r_e_min, g_r_e_max)
            & r_z_e.between(r_z_e_min, r_z_e_max)
        )

        # We want to switch between psfmags and fibermags depending on
        # ls.type parameter (PSF or extended)
        # For 'PSF' targets, we use psfmags, but for extended sources use fiber2mags
        opt_prov = peewee.Case(
            None,
            (
                ((ls.type == "PSF") & valid_p, "sdss_psfmag_from_lsdr10"),
                ((ls.type != "PSF") & valid_e, "sdss_fiber2mag_from_lsdr10"),
            ),
            "undefined",
        )

        magnitude_g = peewee.Case(
            None,
            (
                ((ls.type == "PSF") & valid_p, g_p.cast("float")),
                ((ls.type != "PSF") & valid_e, g_e.cast("float")),
            ),
            "NaN",
        )
        magnitude_r = peewee.Case(
            None,
            (
                ((ls.type == "PSF") & valid_p, r_p.cast("float")),
                ((ls.type != "PSF") & valid_e, r_e.cast("float")),
            ),
            "NaN",
        )
        magnitude_i = peewee.Case(
            None,
            (
                ((ls.type == "PSF") & valid_p, i_p.cast("float")),
                ((ls.type != "PSF") & valid_e, i_e.cast("float")),
            ),
            "NaN",
        )
        magnitude_z = peewee.Case(
            None,
            (
                ((ls.type == "PSF") & valid_p, z_p.cast("float")),
                ((ls.type != "PSF") & valid_e, z_e.cast("float")),
            ),
            "NaN",
        )

        magnitude_gaia_g = peewee.Case(
            None, ((ls.gaia_phot_g_mean_mag.between(0.1, 29.9), ls.gaia_phot_g_mean_mag),), "NaN"
        )
        magnitude_gaia_bp = peewee.Case(
            None, ((ls.gaia_phot_bp_mean_mag.between(0.1, 29.9), ls.gaia_phot_bp_mean_mag),), "NaN"
        )
        magnitude_gaia_rp = peewee.Case(
            None, ((ls.gaia_phot_rp_mean_mag.between(0.1, 29.9), ls.gaia_phot_rp_mean_mag),), "NaN"
        )

        query = (
            c.select(
                c.catalogid.alias("catalogid"),
                priority.alias("priority"),
                value.alias("value"),
                cadence.alias("cadence"),
                instrument.alias("instrument"),
                opt_prov.alias("optical_prov"),
                magnitude_g.alias("g"),
                magnitude_r.alias("r"),
                magnitude_i.alias("i"),
                magnitude_z.alias("z"),
                magnitude_gaia_g.alias("gaia_g"),
                magnitude_gaia_bp.alias("bp"),
                magnitude_gaia_rp.alias("rp"),
                ls.ls_id.alias("ls_id"),  # extra
                ls.gaia_dr3_source_id.alias("gaia_dr3_source_id"),  # extra
                ls.gaia_dr2_source_id.alias("gaia_dr2_source_id"),  # extra
                x.ero_detuid.alias("ero_detuid"),  # extra
                x.ero_flux.alias("ero_flux"),  # extra
                x.ero_det_like.alias("ero_det_like"),  # extra
                x.ero_flags.alias("ero_flags"),  # extra
                x.xmatch_flags.alias("xmatch_flags"),  # extra
                x.xmatch_metric.alias("xmatch_metric"),  # extra
                c.ra.alias("ra"),  # extra
                c.dec.alias("dec"),  # extra
                g0_p.alias("ls10_mag_g"),  # extra
                r0_p.alias("ls10_mag_r"),  # extra
                i0_p.alias("ls10_mag_i"),  # extra
                z0_p.alias("ls10_mag_z"),  # extra
                g0_e.alias("ls10_fibermag_g"),  # extra
                r0_e.alias("ls10_fibermag_r"),  # extra
                i0_e.alias("ls10_fibermag_i"),  # extra
                z0_e.alias("ls10_fibermag_z"),  # extra
                ls.nobs_g.alias("ls10_nobs_g"),
                ls.nobs_r.alias("ls10_nobs_r"),
                ls.nobs_i.alias("ls10_nobs_i"),
                ls.nobs_z.alias("ls10_nobs_z"),
                ls.type.alias("ls10_type"),  # extra
                ls.shape_r.alias("ls10_shape_r"),  # extra
                ls.ref_cat.alias("ls_ref_cat"),  # extra
                ls.ref_id.alias("ls_ref_id"),  # extra
                ls.maskbits.alias("ls_maskbits"),  # extra
                ls.fitbits.alias("ls_fitbits"),  # extra
                # gal_lat.alias('abs_gal_lat'),  # extra
            )
            .join(c2ls)
            .join(ls)
            .join(x, on=(ls.ls_id == x.ls_id))
            # admin criteria
            .where(
                c.version_id == version_id,
                c2ls.version_id == version_id,
                c2ls.best >> True,
            )
            # science criteria
            .where(
                (x.ero_version == self.parameters["ero_version"]),
                (x.xmatch_method == self.parameters["xmatch_method"]),
                (x.xmatch_version == self.parameters["xmatch_version"]),
                (x.opt_cat == self.parameters["opt_cat"]),
                (
                    (ls.fiberflux_r.between(fiberflux_r_min, fiberflux_r_max))
                    | (ls.fiberflux_i.between(fiberflux_i_min, fiberflux_i_max))
                    | (ls.fiberflux_z.between(fiberflux_z_min, fiberflux_z_max))
                ),
                # gaia safety checks to avoid bad ls photometry
                ~(ls.gaia_phot_g_mean_mag.between(0.1, self.parameters["gaia_g_mag_limit"])),
                ~(ls.gaia_phot_rp_mean_mag.between(0.1, self.parameters["gaia_rp_mag_limit"])),
            )
            # .group_by(ls)   # avoid duplicates - we trust the legacy survey entries
            .distinct([ls.ls_id])  # avoid duplicates - we trust the legacy survey entries
        )

        if self.only_faintest_cadence:
            query = query.where(cadence == cadence3)

        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    c.ra, c.dec, query_region[0], query_region[1], query_region[2]
                )
            )

        return query


# ##################################################################################
class BhmSpidersAgnTdaD3Carton(BhmSpidersAgnTdaCarton):
    name = "bhm_spiders_agn_tda_d3"
    only_faintest_cadence = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2021-01-17
# @Filename: mwm_yso.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import MWM_TESS_OB, Catalog, CatalogToTIC_v8, Gaia_DR2, TIC_v8

from target_selection.cartons import BaseCarton


# See catalog.py for the name of peewee model names corresponding
# to postgres table names:
# https://github.com/sdss/sdssdb/blob/master/python/sdssdb/peewee/sdss5db/catalogdb.py
#
# For example
# peewee Model name ---> postgres table name
# Gaia_DR2 --->'catalogdb.gaia_dr2_source'
# MWM_TESS_OB --->'catalogdb.mwm_tess_ob'


class MWM_TESS_OB_Carton(BaseCarton):
    """MWM TESS OB
    Owner: Nathan De Lee

    Shorthand name: mwm_tess_ob

    Mock Target catalogs with GAIA DR2 ids for the TESS
    CVZ OBAF stars

    Simplified Description of selection criteria:
    a random subsample of TIC targets in both TESS CVZs,
    selected according to type of their variability
    (i.e., if (var_type = eclipsing_binary .and. var_type = pulsator
     .and. coord = TESS_CVZ) select )

    Wiki page:

    Additional source catalogs needed:
    There necessary variability information is not currently available,
    so we are going to use mock targets from the following
    two lists of Gaia DR2 IDs:

    mwm_tess_ob_8x1_GaiaID.csv
    mwm_tess_ob_8x3_GaiaID.csv

    Additional cross-matching needed: None

    return columns:  Gaia ID, Hmag

    cadence options for these targets
    (list all options, even though no single target will receive more than one):
    There are three cadences that will be used by these targets.

    For all targets in mwm_tess_ob_8x1_GaiaID.csv use
        instrument =  "APOGEE"
        cadence = "bright_8x1"

    For targets in mwm_tess_ob_8x3_GaiaID.csv
    if H < 11 then use
        instrument = "APOGEE"
        cadence = "bright_8x2"

    if H > =11 then use
        instrument = "APOGEE"
        cadence = "bright_8x4"

    Pseudo SQL (optional):

    Implementation:

    Non-SQL implementation:

    Lead Contact: Andrew Tkachenko
    """

    name = "mwm_tess_ob"
    category = "science"
    instrument = None  # assigned in query
    cadence = None  # assigned in query
    program = "mwm_tessob"
    mapper = "MWM"
    priority = 2200  # from the table in the V0.5 carton wiki page

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToTIC_v8.select(
                CatalogToTIC_v8.catalogid,
                Gaia_DR2.source_id,
                Gaia_DR2.ra.alias("gaia_dr2_ra"),
                Gaia_DR2.dec.alias("gaia_dr2_dec"),
                Gaia_DR2.phot_g_mean_mag.alias("gaia_dr2_g"),
                Gaia_DR2.phot_bp_mean_mag,
                Gaia_DR2.phot_rp_mean_mag,
                Gaia_DR2.parallax,
                MWM_TESS_OB.ra,
                MWM_TESS_OB.dec,
                MWM_TESS_OB.h_mag,
                MWM_TESS_OB.instrument,
                MWM_TESS_OB.cadence,
            )
            .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
            .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
            .join(MWM_TESS_OB, on=(Gaia_DR2.source_id == MWM_TESS_OB.gaia_dr2_id))
            .where(CatalogToTIC_v8.version_id == version_id, CatalogToTIC_v8.best >> True)
        )

        if query_region:
            query = query.join_from(CatalogToTIC_v8, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Tom Dwelly
# @Date: 2021-11-02
# @Filename: bhm_galaxies.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import numpy
import pandas
import peewee

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToLegacy_Survey_DR10,
    Legacy_Survey_DR10,
)

from target_selection.cartons.base import BaseCarton
from target_selection.mag_flux import AB2nMgy


"""
Details: Start here
https://wiki.sdss.org/display/OPS/Defining+target+selection+and+cadence+algorithms
  see particularly: https://wiki.sdss.org/display/BHM/BHM+Cartons+of+Last+Resort

This module provides the following BHM cartons:
bhm_colr_galaxies_lsdr10
bhm_colr_galaxies_lsdr10_d3

"""


# used by cartons that need to compute Galactic latitude:
north_gal_pole_ra = 192.85948  # deg, J2000
north_gal_pole_dec = +27.12825  # deg, J2000


class BhmColrGalaxiesLsdr10Carton(BaseCarton):
    """
    A sample of bright galaxies selected from legacysurvey/dr10
    photometry+astrometry+morphology
    """

    name = "bhm_colr_galaxies_lsdr10"
    category = "science"
    mapper = "BHM"
    program = "bhm_filler"
    tile = False
    # we do not set cadence here since
    # cadence is set later below
    # cadence = 'dark_1x1'
    instrument = "BOSS"
    can_offset = False
    only_faintest_cadence = False

    def build_query(self, version_id, query_region=None):
        c = Catalog.alias()
        c2ls = CatalogToLegacy_Survey_DR10.alias()
        ls = Legacy_Survey_DR10.alias()

        # set the Carton priority+values here - read from yaml
        priority = peewee.Value(int(self.parameters.get("priority", 10000)))
        value = peewee.Value(self.parameters.get("value", 0.0)).cast("float")
        inertial = peewee.Value(True)
        instrument = peewee.Value(self.instrument)
        # cadence = peewee.Value(self.parameters.get('cadence', self.cadence))

        # dered_flux_z_min = AB2nMgy(self.parameters["dered_mag_z_max"])  # moved to post_process
        # dered_fiberflux_z_min = AB2nMgy(self.parameters["dered_fibermag_z_max"])  # ditto
        fiberflux_z_min = AB2nMgy(self.parameters["fibermag_z_max"])
        fiberflux_z_max = AB2nMgy(self.parameters["fibermag_z_min"])
        fiberflux_r_min = AB2nMgy(self.parameters["fibermag_r_max"])
        fiberflux_r_max = AB2nMgy(self.parameters["fibermag_r_min"])
        fiberflux_g_min = AB2nMgy(self.parameters["fibermag_g_max"])
        fiberflux_g_max = AB2nMgy(self.parameters["fibermag_g_min"])

        fiberflux_r_min_for_cadence1 = AB2nMgy(self.parameters["fibermag_r_for_cadence1"])
        fiberflux_r_min_for_cadence2 = AB2nMgy(self.parameters["fibermag_r_for_cadence2"])

        # compute transformed SDSS mags uniformly
        # transform the legacysurvey grz into sdss fiber2mag griz
        # https://wiki.sdss.org/display/BHM/BHM+magnitude+transformations+for+v0.5

        # extract coeffs from fit logs via:
        # awk 'BEGIN {print("coeffs = {")} /POLYFIT/{ if($3~/sdss_psfmag/){pe="p"} else if ($3~/sdss_fiber2mag/){pe="e"} else{pe="error"}; printf("\"%s%d_%s\": %s,\n", substr($3,length($3)), $8, pe, $10)} END {print("}")}'  bhm_colr_galaxies_lsdr8/lsdr8_fibermag_to_sdss_fiber2mag_?_results.log   # noqa

        coeffs = {
            "g2_e": -0.611168,
            "g1_e": 1.637641,
            "g0_e": -0.684489,
            "i2_e": -1.821603,
            "i1_e": 2.410118,
            "i0_e": -0.895875,
            "r2_e": -0.337699,
            "r1_e": 1.038298,
            "r0_e": -0.371602,
            "z2_e": -1.609400,
            "z1_e": 2.678952,
            "z0_e": -0.821672,
        }

        nMgy_min = 1e-3  # equiv to AB=30
        # extended - start from ls10 fiberfluxes
        # TODO investigate using real lsdr10 i-band where available
        g0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_g))
        r0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_r))
        i0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_i))
        z0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_z))
        g_r_e = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.fiberflux_g)
            / peewee.fn.greatest(nMgy_min, ls.fiberflux_r)
        )
        r_z_e = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.fiberflux_r)
            / peewee.fn.greatest(nMgy_min, ls.fiberflux_z)
        )

        g_e = g0_e + coeffs["g0_e"] + coeffs["g1_e"] * g_r_e + coeffs["g2_e"] * g_r_e * g_r_e
        r_e = r0_e + coeffs["r0_e"] + coeffs["r1_e"] * g_r_e + coeffs["r2_e"] * g_r_e * g_r_e
        i_e = r0_e + coeffs["i0_e"] + coeffs["i1_e"] * r_z_e + coeffs["i2_e"] * r_z_e * r_z_e
        z_e = z0_e + coeffs["z0_e"] + coeffs["z1_e"] * r_z_e + coeffs["z2_e"] * r_z_e * r_z_e

        valid_colour_min_g_r = 0.0
        valid_colour_max_g_r = 2.0
        valid_colour_min_r_z = 0.2
        valid_colour_max_r_z = 1.2

        # select_flux_ratio_min_g_r = 10**(-0.4 * self.parameters['select_max_g_r'])
        # select_flux_ratio_max_g_r = 10**(-0.4 * self.parameters['select_min_g_r'])
        # select_flux_ratio_min_r_z = 10**(-0.4 * self.parameters['select_max_r_z'])
        # select_flux_ratio_max_r_z = 10**(-0.4 * self.parameters['select_min_r_z'])

        # magnitude validity checks
        valid = (
            g0_e.between(0.1, 29.9)
            & r0_e.between(0.1, 29.9)
            & z0_e.between(0.1, 29.9)
            & g_r_e.between(valid_colour_min_g_r, valid_colour_max_g_r)
            & r_z_e.between(valid_colour_min_r_z, valid_colour_max_r_z)
        )

        opt_prov = peewee.Case(None, ((valid, "sdss_fiber2mag_from_lsdr10"),), "undefined")
        magnitude_g = peewee.Case(None, ((valid, g_e),), "NaN")
        magnitude_r = peewee.Case(None, ((valid, r_e),), "NaN")
        magnitude_i = peewee.Case(None, ((valid, i_e),), "NaN")
        magnitude_z = peewee.Case(None, ((valid, z_e),), "NaN")
        magnitude_gaia_g = peewee.Case(
            None, ((ls.gaia_phot_g_mean_mag.between(0.1, 29.9), ls.gaia_phot_g_mean_mag),), "NaN"
        )
        magnitude_gaia_bp = peewee.Case(
            None, ((ls.gaia_phot_bp_mean_mag.between(0.1, 29.9), ls.gaia_phot_bp_mean_mag),), "NaN"
        )
        magnitude_gaia_rp = peewee.Case(
            None, ((ls.gaia_phot_rp_mean_mag.between(0.1, 29.9), ls.gaia_phot_rp_mean_mag),), "NaN"
        )

        opt_prov = peewee.Case(None, ((valid, "sdss_fiber2mag_from_lsdr10"),), "undefined")

        cadence1 = self.parameters["cadence1"]
        cadence2 = self.parameters["cadence2"]
        cadence3 = self.parameters["cadence3"]
        cadence = peewee.Case(
            None,
            (
                (ls.fiberflux_r > fiberflux_r_min_for_cadence1, cadence1),
                (ls.fiberflux_r > fiberflux_r_min_for_cadence2, cadence2),
            ),
            cadence3,
        )

        # compute the abs(Galactic latitude):
        # we return this in the query result, and then and test aginst it in post_process
        gal_lat = peewee.fn.abs(
            90.0 - peewee.fn.q3c_dist(north_gal_pole_ra, north_gal_pole_dec, c.ra, c.dec)
        )

        # https://www.legacysurvey.org/dr10/bitmasks/
        maskbits_mask = 2**1 + 2**13
        # maskbits_mask = 2**0 + 2**1 + 2**4 + 2**7 + 2**8 + 2**10 + 2**11 + 2**13
        # fitbits_mask = 2**5 + 2**6 + 2**7 + 2**8 + 2**12
        query = (
            c.select(
                c.catalogid.alias("catalogid"),
                ls.ls_id.alias("ls_id"),  # extra
                c.ra.alias("ra"),  # extra
                c.dec.alias("dec"),  # extra
                priority.alias("priority"),
                value.cast("real").alias("value"),
                cadence.alias("cadence"),
                instrument.alias("instrument"),
                opt_prov.alias("optical_prov"),
                magnitude_g.cast("real").alias("g"),
                magnitude_r.cast("real").alias("r"),
                magnitude_i.cast("real").alias("i"),
                magnitude_z.cast("real").alias("z"),
                magnitude_gaia_g.alias("gaia_g"),
                magnitude_gaia_bp.alias("bp"),
                magnitude_gaia_rp.alias("rp"),
                inertial.alias("inertial"),
                g0_e.cast("real").alias("ls10_fibermag_g"),  # extra
                r0_e.cast("real").alias("ls10_fibermag_r"),  # extra
                i0_e.cast("real").alias("ls10_fibermag_i"),  # extra
                z0_e.cast("real").alias("ls10_fibermag_z"),  # extra
                g_r_e.cast("real").alias("ls10_g_r_e"),  # extra
                r_z_e.cast("real").alias("ls10_r_z_e"),  # extra
                ls.flux_g.alias("ls10_flux_g"),  # extra
                ls.flux_r.alias("ls10_flux_r"),  # extra
                ls.flux_i.alias("ls10_flux_i"),  # extra
                ls.flux_z.alias("ls10_flux_z"),  # extra
                ls.fiberflux_z.alias("ls10_fiberflux_z"),  # extra
                ls.ebv.alias("ls10_ebv"),  # extra
                ls.mw_transmission_z.alias("ls10_mw_transmission_z"),  # extra
                ls.shape_r.alias("ls10_shape_r"),  # extra
                ls.maskbits.alias("ls10_maskbits"),  # extra
                ls.fitbits.alias("ls10_fitbits"),  # extra
                gal_lat.alias("abs_gal_lat"),  # extra
            )
            .join(c2ls)
            .join(ls)
            .where(
                c.version_id == version_id,
                c2ls.version_id == version_id,
                ls.type != "PSF",
                ls.parallax <= 0.0,
                # ls.flux_z > dered_flux_z_min * ls.mw_transmission_z,  # moved to post_process
                # ls.fiberflux_z > dered_fiberflux_z_min * ls.mw_transmission_z,  # ditto
                ls.fiberflux_g > fiberflux_g_min,
                ls.fiberflux_r > fiberflux_r_min,
                ls.fiberflux_z > fiberflux_z_min,
                ls.fiberflux_g < fiberflux_g_max,
                ls.fiberflux_r < fiberflux_r_max,
                ls.fiberflux_z < fiberflux_z_max,
                # safety check using Gaia mags to avoid bad ls photometry
                ~(ls.gaia_phot_g_mean_mag.between(0.1, self.parameters["gaia_g_mag_limit"])),
                ~(ls.gaia_phot_rp_mean_mag.between(0.1, self.parameters["gaia_rp_mag_limit"])),
                ls.shape_r >= self.parameters["shape_r_min"],
                # gal_lat > self.parameters["min_gal_lat"],  # moved to post_process
                ls.ebv < self.parameters["max_ebv"],
                (ls.maskbits.bin_and(maskbits_mask) == 0),  # avoid bad ls data
                # (ls.fitbits.bin_and(fitbits_mask) == 0),  # avoid bad ls fits
            )
            # .where(c.catalogid.between(63050396500000000,   # debug
            #                            63050396550000000))  # debug
            .distinct(c.catalogid)
        )

        if self.only_faintest_cadence:
            query = query.where(ls.fiberflux_r <= fiberflux_r_min_for_cadence2)

        # query_region[0] is ra of center of the region, degrees
        # query_region[1] is dec of center of the region, degrees
        # query_region[2] is radius of the region, degrees
        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    c.ra, c.dec, query_region[0], query_region[1], query_region[2]
                )
            )

        return query

    def post_process(self, model, **kwargs):
        """Runs post-process."""
        dered_flux_z_min = AB2nMgy(self.parameters["dered_mag_z_max"])
        dered_fiberflux_z_min = AB2nMgy(self.parameters["dered_fibermag_z_max"])

        data = pandas.read_sql(
            (
                "SELECT catalogid,abs_gal_lat,ls10_flux_z,"
                "ls10_fiberflux_z,ls10_mw_transmission_z "
                f"from {self.path}"
            ),
            self.database,
        )

        valid = numpy.where(
            (data["abs_gal_lat"] > self.parameters["min_gal_lat"])
            & (data["ls10_flux_z"] > dered_flux_z_min * data["ls10_mw_transmission_z"])
            & (data["ls10_fiberflux_z"] > dered_fiberflux_z_min * data["ls10_mw_transmission_z"]),
            True,
            False,
        )

        print(
            "During post-processing we down-selected "
            f"{numpy.count_nonzero(valid)}/{len(data)} rows"
        )

        data = data[valid]

        valid_cids = data.catalogid.values

        # This way seems faster than updating from a list of values.
        values_cids = ",".join(f"({vc})" for vc in valid_cids)
        self.database.execute_sql(
            "CREATE TEMP TABLE valid_cids AS SELECT * "
            f"FROM (VALUES {values_cids}) "
            "AS t (catalogid)"
        )
        self.database.execute_sql("CREATE INDEX ON valid_cids (catalogid);")
        self.database.execute_sql(f"UPDATE {self.path} SET selected = false")
        self.database.execute_sql(
            f"UPDATE {self.path} SET selected = true "
            "FROM valid_cids vc "
            f"WHERE {self.path}.catalogid = vc.catalogid"
        )

        return super().post_process(model, **kwargs)


class BhmColrGalaxiesLsdr10D3Carton(BhmColrGalaxiesLsdr10Carton):
    name = "bhm_colr_galaxies_lsdr10_d3"
    only_faintest_cadence = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2022-12-11
# @Filename: mwm_tess_planets.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import numpy
import peewee

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToTIC_v8,
    TESS_TOI_v1,
    TIC_v8,
    TwoMassPSC,
)

from target_selection.cartons import BaseCarton


def h2exp(hmag, sn=100, exptime=15.0):
    """
    This function takes in a hmag and given signal to noise and spits back
    the required time. Based on Hmag = 11 at S/N 100 in an hour.
    """
    # Scale the hmag based on t = (1 hour)*10^(0.4*(H-11))
    # Then I cut it up into 15 minute exposures.
    time = 60 * (sn**2 / 100.0**2) * 10 ** (0.4 * (hmag - 11))
    nexp = numpy.array(numpy.round(time / exptime))
    # Min value is 1
    nexp[(nexp == 0)] = 1
    # Set Nan's to nan
    nexp[numpy.isnan(hmag)] = numpy.nan
    return nexp


class MWM_TESS_2min_apogee_Carton(BaseCarton):
    """TESS Planets.
    Definition:
        All planet-candidate TOIs + Community TOIs (CTIOs), TESS
        Candidate-Target List (CTL) selected by a TIC Priority cut.
        All cut between 7 < H < 12.

    All the candidates are compiled in catalogdb.tess_toi_v1. Note that this list
    contains duplicates on ticid.

    We changed the carton class name, the variable 'name'
    and the temp table name in post_process().

    The old name of this cartoon is 'mwm_tess_planet'.
    """

    name = "mwm_tess_2min_apogee"
    program = "mwm_planet"
    category = "science"
    mapper = "MWM"
    instrument = "APOGEE"
    cadence = None  # cadence is set in post_process()
    priority = 2810
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = (
            TESS_TOI_v1.select(
                CatalogToTIC_v8.catalogid,
                TESS_TOI_v1.ticid,
                TESS_TOI_v1.tess_disposition,
                TwoMassPSC.h_m.alias("hmag"),
                TESS_TOI_v1.target_type.alias("tess_target_type"),
            )
            .join(TIC_v8, on=(TESS_TOI_v1.ticid == TIC_v8.id))
            .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
            .switch(TIC_v8)
            .join(CatalogToTIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
            .where(
                CatalogToTIC_v8.version_id == version_id,
                CatalogToTIC_v8.best >> True,
                TwoMassPSC.h_m > 7,
                TwoMassPSC.h_m < 12,
            )
            .distinct([TESS_TOI_v1.ticid])
        )

        if query_region:
            query = query.join_from(CatalogToTIC_v8, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query

    def post_process(self, model):
        cursor = self.database.execute_sql(
            "select catalogid, hmag from " + " sandbox.temp_mwm_tess_2min_apogee ;"
        )

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_hmag = float(output[i][1])

            numexp = h2exp(current_hmag, sn=80)

            # for numexp == 1, we use bright_1x1 since
            # there is no cadence bright_flexible_1x1
            if numexp == 1:
                current_cadence = "bright_1x1"
            else:
                current_cadence = "bright_flexible_" + str(int(numexp)) + "x1"

            self.database.execute_sql(
                " update sandbox.temp_mwm_tess_2min_apogee "
                + " set cadence = '"
                + current_cadence
                + "'"
                " where catalogid = " + str(current_catalogid) + ";"
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-06-02
# @Filename: mwm_snc.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import random

import peewee
from peewee import fn

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToGaia_DR3,
    CatalogToTwoMassPSC,
    Gaia_DR3,
    TwoMassPSC,
)

from target_selection.cartons import BaseCarton


class MWM_SNC_100pc_Carton(BaseCarton):
    """MWM Solar Neighbourhood Census 100pc.

    Definition:

        - Divide the sky in crowded/not crowded regions via simple geometric
          cuts in (l,b).
        - Select all stars that fall within their one-sigma uncertainty within
          100pc in the not crowded regions.
        - Select all stars that fall within their one-sigma uncertainty
          within 100pc and have an astrometric_excess_noise < 2 in the
          crowded regions.

    Pseudo SQL:

        parallax-parallax_error>10 &&
        ( (astrometric_excess_noise < 2 &&
           ( (l <= 180 && b < -0.139 * l + 25 && b > 0.139 * l - 25) ||
             (l > 180 && b > -0.139 * l + 25 && b < 0.139 * l - 25) ||
             (sqrt(pow(l - 303.2, 2) + 2 * pow(b + 44.4, 2)) < 5) ||
             (sqrt(pow(l - 280.3, 2) + 2 * pow(b + 33.0, 2)) < 8) )
        ) ||
        !( (l <= 180 && b < -0.139 * l + 25 && b > 0.139 * l - 25) ||
           (l > 180 && b > -0.139 * l + 25 && b < 0.139 * l - 25) ||
           (sqrt(pow(l - 303.2, 2) + 2 * pow(b + 44.4, 2)) < 5) ||
           (sqrt(pow(l - 280.3, 2) + 2 * pow(b + 33.0, 2)) < 8) )
        )

    This is a base carton. Actual cartons are implemented as subclasses for the
    different magnitude cuts.

    """

    def build_query(self, version_id, query_region=None):
        ll = Gaia_DR3.l
        bb = Gaia_DR3.b

        # Dense regions (Galactic plane, SMC, LMC).
        gal_cut = (
            ((ll <= 180) & (bb < (-0.139 * ll + 25)) & (bb > (0.139 * ll - 25)))
            | ((ll > 180) & (bb > (-0.139 * ll + 25)) & (bb < (0.139 * ll - 25)))
            | (fn.sqrt(fn.pow(ll - 303.2, 2) + 2 * fn.pow(bb + 44.4, 2)) < 5)
            | (fn.sqrt(fn.pow(ll - 280.3, 2) + 2 * fn.pow(bb + 33.0, 2)) < 8)
        )

        cte = (
            Gaia_DR3.select(Gaia_DR3.source_id)
            .where(Gaia_DR3.parallax > 0, (Gaia_DR3.parallax - Gaia_DR3.parallax_error) > 10)
            .cte("plx_cte", materialized=True)
        )

        query = (
            Gaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra,
                Gaia_DR3.dec,
                Gaia_DR3.phot_g_mean_mag,
                ll,
                bb,
            )
            .join(CatalogToGaia_DR3)
            .join_from(Gaia_DR3, cte, on=(Gaia_DR3.source_id == cte.c.source_id))
            .where(((Gaia_DR3.astrometric_excess_noise < 2) & gal_cut) | ~gal_cut)
            .where(CatalogToGaia_DR3.version_id == version_id, CatalogToGaia_DR3.best >> True)
            .with_cte(cte)
        )

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_SNC_100pc_APOGEE_Carton(MWM_SNC_100pc_Carton):
    """MWM SNC 100pc targets to be observed with APOGEE."""

    name = "mwm_snc_100pc_apogee"
    category = "science"
    program = "mwm_snc"
    mapper = "MWM"
    instrument = "APOGEE"
    cadence = "bright_1x1"
    priority = 2705
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region=query_region)
        query = (
            query.switch(CatalogToGaia_DR3)
            .join(
                CatalogToTwoMassPSC,
                on=(CatalogToGaia_DR3.catalogid == CatalogToTwoMassPSC.catalogid),
            )
            .join(TwoMassPSC)
            .where(
                CatalogToTwoMassPSC.version_id == version_id,
                CatalogToTwoMassPSC.best >> True,
                TwoMassPSC.h_m < 11,
            )
        )

        return query


class MWM_SNC_100pc_BOSS_Carton(MWM_SNC_100pc_Carton):
    """MWM SNC 100pc targets to be observed with BOSS.

    Shorthand name:  mwm_snc_100pc_boss
    https://sdss-wiki.atlassian.net/wiki/spaces/OPS/pages/13707660/Cartons+for+post+v1

    Existing carton code:
    https://github.com/sdss/target_selection/blob/main/python/target_selection/cartons/mwm_snc.py

    Simplified Description of selection criteria :
    Select everything within 100pc (i.e. parallax > 10mas) and Gaia-G
    fainter than 10th mag. Add some constraints on astrometric excess noise
    in crowded regions (Galactic plane, LMC, SMC, towards Galactic center).
    For details, see here. Pseudo-code: For the pseduo-code, I will shorten
    it and use the definition of "gal_cut" from the existing carton code.
    Also, the red text below is what is being added to the target selection
    compared to the original v1 carton.
    (parallax > 0) & (parallax - parallax_error > 10) & (phot_g_mean_mag > 10) &
    ((astrometric_excess_noise < 2) & (ll >= 20) & (ll <= 340) & gal_cut) |
    ((astrometric_excess_noise < 2) & (ruwe < 1.2) &
    ((1 + 0.015 * (phot_bp_mean_mag - phot_rp_mean_mag)^2) < phot_bp_rp_excess_factor) &
    (phot_bp_rp_excess_factor <
    (1.3 + 0.06 * (phot_bp_mean_mag - phot_rp_mean_mag)^2)) &
    ((ll < 20) | (ll > 340)) & gal_cut) | ~gal_cut)

    Return columns: Same columns as before + astrometric_excess_noise,
    phot_bp_rp_excess_factor, ruwe, phot_bp_mean_mag , phot_rp_mean_mag

    Metadata:
    Priority: 1800
    Cadence:  if G < 16, bright_2x1 ; if G > 16, dark_2x1
    Instrument: BOSS
    can_offset = True
    Lead contact: Ilija Medan

    """

    name = "mwm_snc_100pc_boss"
    category = "science"
    program = "mwm_snc"
    mapper = "MWM"
    instrument = "BOSS"
    cadence = None  # cadence is set in post_process()
    priority = 1800
    can_offset = True

    def build_query(self, version_id, query_region=None):
        ll = Gaia_DR3.l
        bb = Gaia_DR3.b

        # Dense regions (Galactic plane, SMC, LMC).
        gal_cut = (
            ((ll <= 180) & (bb < (-0.139 * ll + 25)) & (bb > (0.139 * ll - 25)))
            | ((ll > 180) & (bb > (-0.139 * ll + 25)) & (bb < (0.139 * ll - 25)))
            | (fn.sqrt(fn.pow(ll - 303.2, 2) + 2 * fn.pow(bb + 44.4, 2)) < 5)
            | (fn.sqrt(fn.pow(ll - 280.3, 2) + 2 * fn.pow(bb + 33.0, 2)) < 8)
        )

        query = super().build_query(version_id, query_region)
        query = query.where(
            Gaia_DR3.parallax > 0,
            Gaia_DR3.parallax - Gaia_DR3.parallax_error > 10,
            Gaia_DR3.phot_g_mean_mag > 10,
            ((Gaia_DR3.astrometric_excess_noise < 2) & (ll >= 20) & (ll <= 340) & gal_cut)
            | (
                (Gaia_DR3.astrometric_excess_noise < 2)
                & (Gaia_DR3.ruwe < 1.2)
                & (
                    (1 + 0.015 * fn.pow(Gaia_DR3.phot_bp_mean_mag - Gaia_DR3.phot_rp_mean_mag, 2))
                    < Gaia_DR3.phot_bp_rp_excess_factor
                )
                & (
                    (
                        Gaia_DR3.phot_bp_rp_excess_factor
                        < (
                            1.3
                            + 0.06
                            * fn.pow(Gaia_DR3.phot_bp_mean_mag - Gaia_DR3.phot_rp_mean_mag, 2)
                        )
                    )
                    & ((ll < 20) | (ll > 340))
                    & gal_cut
                )
                | ~gal_cut
            ),
        )

        return query

    def post_process(self, model, **kwargs):
        # G < 16 => cadence = bright_2x1
        model.update(cadence="bright_2x1").where(model.phot_g_mean_mag < 16).execute()

        # G > 16 => cadence = dark_2x1
        model.update(cadence="dark_2x1").where(model.phot_g_mean_mag >= 16).execute()

        return model


class MWM_SNC_100pc_BOSS_single_Carton(MWM_SNC_100pc_BOSS_Carton):
    """MWM SNC 100pc targets to be observed with BOSS.

    Shorthand name:  mwm_snc_100pc_boss_single
    https://sdss-wiki.atlassian.net/wiki/spaces/OPS/pages/13707660/Cartons+for+post+v1
    Simplified Description of selection criteria :
    NOTE this is a subcarton of mwm_snc_100pc_boss (see above) where the only change
    is the cadence and priority.

    Metadata:
    Priority: 1806
    Cadence:  if G < 16, bright_1x1 ; if G > 16, dark_1x1
    Instrument: BOSS
    can_offset = True
    Lead contact: Ilija Medan
    """

    name = "mwm_snc_100pc_boss_single"
    category = "science"
    program = "mwm_snc"
    mapper = "MWM"
    instrument = "BOSS"
    cadence = None  # cadence is set in post_process()
    priority = 1806
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        return query

    def post_process(self, model, **kwargs):
        # G < 16 => cadence = bright_1x1
        model.update(cadence="bright_1x1").where(model.phot_g_mean_mag < 16).execute()

        # G > 16 => cadence = dark_1x1
        model.update(cadence="dark_1x1").where(model.phot_g_mean_mag >= 16).execute()

        return model


class MWM_SNC_Ext_Carton(BaseCarton):
    """MWM Solar Neighbourhood Census extension.

    Shorthand name: mwm_snc_ext_apogee
    Simplified Description of selection criteria:
        Above 100pc, apply a linear cut in absolute G-mag and distance to select increasingly
        more massive stars at larger distances. 2MASS H-band brighter than 11mag.
        Add some constraints on astrometric excess noise in crowded regions
        (Galactic plane, LMC, SMC).
    Pseudo-code:
        ((parallax<10) && (phot_g_mean_mag+5*log10(parallax/1000)+5) <
          19.125-0.03225*(1000/parallax))  &&
          ( (astrometric_excess_noise < 2 &&
            ( (l<=180 && b<-0.139*l + 25 && b>0.139*l - 25) ||
              (l>180 && b>-0.139*l + 25 && b <0.139*l - 25) ||
              (sqrt( pow(l-303.2,2) + 2*pow(b+44.4,2) ) < 5) ||
              (sqrt( pow(l-280.3,2) + 2*pow(b+33.0,2) ) < 8))) ||
            !( (l<=180 && b<-0.139*l + 25 && b>0.139*l - 25) ||
               (l>180 && b>-0.139*l + 25 && b <0.139*l - 25) ||
               (sqrt( pow(l-303.2,2) + 2*pow(b+44.4,2) ) < 5) ||
               (sqrt( pow(l-280.3,2) + 2*pow(b+33.0,2) ) < 8) ) ) && 2MASS_H < 11


    Shorthand name: mwm_snc_ext_boss
    Simplified Description of selection criteria:
        Above 100pc, apply a linear cut in absolute G-mag and distance to select increasingly
        more massive stars at larger distances. Gaia G-band fainter than 10mag.
        Add some constraints on astrometric excess noise in crowded regions
        (Galactic plane, LMC, SMC).
    Pseudo-code:
        ((parallax<10) && (phot_g_mean_mag+5*log10(parallax/1000)+5) <
          19.125-0.03225*(1000/parallax))  &&
          ( (astrometric_excess_noise < 2 &&
            ( (l<=180 && b<-0.139*l + 25 && b>0.139*l - 25) ||
              (l>180 && b>-0.139*l + 25 && b <0.139*l - 25) ||
              (sqrt( pow(l-303.2,2) + 2*pow(b+44.4,2) ) < 5) ||
              (sqrt( pow(l-280.3,2) + 2*pow(b+33.0,2) ) < 8))) ||
            !( (l<=180 && b<-0.139*l + 25 && b>0.139*l - 25) ||
               (l>180 && b>-0.139*l + 25 && b <0.139*l - 25) ||
               (sqrt( pow(l-303.2,2) + 2*pow(b+44.4,2) ) < 5) ||
               (sqrt( pow(l-280.3,2) + 2*pow(b+33.0,2) ) < 8) ) ) && phot_g_mag > 10

    """

    def build_query(self, version_id, query_region=None):
        ll = Gaia_DR3.l
        bb = Gaia_DR3.b

        # Dense regions (Galactic plane, SMC, LMC).
        gal_cut = (
            ((ll <= 180) & (bb < (-0.139 * ll + 25)) & (bb > (0.139 * ll - 25)))
            | ((ll > 180) & (bb > (-0.139 * ll + 25)) & (bb < (0.139 * ll - 25)))
            | (fn.sqrt(fn.pow(ll - 303.2, 2) + 2 * fn.pow(bb + 44.4, 2)) < 5)
            | (fn.sqrt(fn.pow(ll - 280.3, 2) + 2 * fn.pow(bb + 33.0, 2)) < 8)
        )

        mag_cut = (
            Gaia_DR3.phot_g_mean_mag + 5 * fn.log(Gaia_DR3.parallax / 1000) + 5
        ) < 19.125 - 0.03225 * (1000 / Gaia_DR3.parallax)

        query = (
            Gaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra,
                Gaia_DR3.dec,
                Gaia_DR3.phot_g_mean_mag,
                ll,
                bb,
            )
            .join(CatalogToGaia_DR3)
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                Gaia_DR3.parallax > 0,
                Gaia_DR3.parallax < 10,
                ((Gaia_DR3.astrometric_excess_noise < 2) & gal_cut) | ~gal_cut,
                mag_cut,
            )
        )

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_SNC_Ext_APOGEE_Base_Carton(MWM_SNC_Ext_Carton):
    """SNC extension for APOGEE. See base carton for details.
    This is a base carton. The derived cartons are
    mwm_snc_ext_main_apogee: priority = 2705
    mwm_snc_ext_filler_apogee: priority = 7200
    """

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region=query_region)
        query = (
            query.switch(CatalogToGaia_DR3)
            .join(
                CatalogToTwoMassPSC,
                on=(CatalogToGaia_DR3.catalogid == CatalogToTwoMassPSC.catalogid),
            )
            .join(TwoMassPSC)
            .where(
                CatalogToTwoMassPSC.version_id == version_id,
                CatalogToTwoMassPSC.best >> True,
                TwoMassPSC.h_m < 11,
            )
        )

        return query


class MWM_SNC_Ext_main_APOGEE_Carton(MWM_SNC_Ext_APOGEE_Base_Carton):
    """See base carton for details."""

    name = "mwm_snc_ext_main_apogee"
    category = "science"
    program = "mwm_snc"
    mapper = "MWM"
    instrument = "APOGEE"
    cadence = "bright_1x1"
    priority = 2705
    can_offset = True

    def post_process(self, model):
        """
        a set random seed, select 1/16 of the targets
        """
        cursor = self.database.execute_sql(
            "update sandbox.temp_mwm_snc_ext_main_apogee " + "set selected = false;"
        )

        # The below "order by catalogid" ensures that the random selection
        # further below gives the same result every time we run this carton.
        cursor = self.database.execute_sql(
            "select catalogid from "
            + " sandbox.temp_mwm_snc_ext_main_apogee "
            + " order by catalogid;"
        )

        output = cursor.fetchall()

        # This random seed must be the same as in mwm_snc_ext_filler_apogee
        random.seed(1234)
        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_random = random.randrange(16)
            if current_random == 0:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_snc_ext_main_apogee "
                    + " set selected = true "
                    + " where catalogid = "
                    + str(current_catalogid)
                    + ";"
                )


class MWM_SNC_Ext_filler_APOGEE_Carton(MWM_SNC_Ext_APOGEE_Base_Carton):
    """See base carton for details."""

    name = "mwm_snc_ext_filler_apogee"
    category = "science"
    program = "mwm_snc"
    mapper = "MWM"
    instrument = "APOGEE"
    cadence = "bright_1x1"
    priority = 7200
    can_offset = True

    def post_process(self, model):
        """
        a set random seed, select 15/16 of the targets
        """
        cursor = self.database.execute_sql(
            "update sandbox.temp_mwm_snc_ext_filler_apogee " + "set selected = false;"
        )

        # The below "order by catalogid" ensures that the random selection
        # further below gives the same result every time we run this carton.
        cursor = self.database.execute_sql(
            "select catalogid from "
            + " sandbox.temp_mwm_snc_ext_filler_apogee "
            + " order by catalogid;"
        )

        output = cursor.fetchall()

        # This random seed must be the same as in mwm_snc_ext_main_apogee
        random.seed(1234)
        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_random = random.randrange(16)
            if current_random != 0:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_snc_ext_filler_apogee "
                    + " set selected = true "
                    + " where catalogid = "
                    + str(current_catalogid)
                    + ";"
                )


class MWM_SNC_Ext_BOSS_Base_Carton(MWM_SNC_Ext_Carton):
    """SNC extension for BOSS. See base carton for details.
    This is a base carton. The derived cartons are
    mwm_snc_ext_main_boss: priority = 2705
    mwm_snc_ext_filler_boss: priority = 7200"""

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(Gaia_DR3.phot_g_mean_mag > 10)

        return query


class MWM_SNC_Ext_main_BOSS_Carton(MWM_SNC_Ext_BOSS_Base_Carton):
    """See base carton for details."""

    # cadence must be None here so that
    # it can be set in post_process().
    # If cadence is not None here then
    # it cannot be set in post_process().
    name = "mwm_snc_ext_main_boss"
    category = "science"
    program = "mwm_snc"
    mapper = "MWM"
    instrument = "BOSS"
    cadence = None  # cadence is set in post_process()
    priority = 2705
    can_offset = True

    def post_process(self, model):
        """
        a set random seed, select 1/20 of the targets
        """

        cursor = self.database.execute_sql(
            "update sandbox.temp_mwm_snc_ext_main_boss " + "set selected = false;"
        )

        # The below "order by catalogid" ensures that the random selection
        # further below gives the same result every time we run this carton.
        cursor = self.database.execute_sql(
            "select catalogid from "
            + " sandbox.temp_mwm_snc_ext_main_boss "
            + " order by catalogid;"
        )

        output = cursor.fetchall()

        # This random seed must be the same as in mwm_snc_ext_filler_boss
        random.seed(5678)
        for i in range(len(output)):
            current_catalogid = output[i][0]

            current_random = random.randrange(20)
            if current_random == 0:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_snc_ext_main_boss "
                    + " set selected = true "
                    + " where catalogid = "
                    + str(current_catalogid)
                    + ";"
                )

            if Gaia_DR3.phot_g_mean_mag < 16:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_snc_ext_main_boss "
                    + " set cadence = 'bright_flexible_2x1' "
                    + " where catalogid = "
                    + str(current_catalogid)
                    + ";"
                )
            else:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_snc_ext_main_boss "
                    + " set cadence = 'dark_flexible_2x1' "
                    + " where catalogid = "
                    + str(current_catalogid)
                    + ";"
                )


class MWM_SNC_Ext_filler_BOSS_Carton(MWM_SNC_Ext_BOSS_Base_Carton):
    """See base carton for details."""

    # cadence must be None here so that
    # it can be set in post_process().
    # If cadence is not None here then
    # it cannot be set in post_process().
    name = "mwm_snc_ext_filler_boss"
    category = "science"
    program = "mwm_snc"
    mapper = "MWM"
    instrument = "BOSS"
    cadence = None  # cadence is set in post_process()
    priority = 7200
    can_offset = True

    def post_process(self, model):
        """
        a set random seed, select 19/20 of the targets
        """

        cursor = self.database.execute_sql(
            "update sandbox.temp_mwm_snc_ext_filler_boss " + "set selected = false;"
        )

        # The below "order by catalogid" ensures that the random selection
        # further below gives the same result every time we run this carton.
        cursor = self.database.execute_sql(
            "select catalogid from "
            + " sandbox.temp_mwm_snc_ext_filler_boss "
            + " order by catalogid;"
        )

        output = cursor.fetchall()

        # This random seed must be the same as in mwm_snc_ext_main_boss
        random.seed(5678)
        for i in range(len(output)):
            current_catalogid = output[i][0]

            current_random = random.randrange(20)
            if current_random != 0:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_snc_ext_filler_boss "
                    + " set selected = true "
                    + " where catalogid = "
                    + str(current_catalogid)
                    + ";"
                )

            if Gaia_DR3.phot_g_mean_mag < 16:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_snc_ext_filler_boss "
                    + " set cadence = 'bright_flexible_2x1' "
                    + " where catalogid = "
                    + str(current_catalogid)
                    + ";"
                )
            else:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_snc_ext_filler_boss "
                    + " set cadence = 'dark_flexible_2x1' "
                    + " where catalogid = "
                    + str(current_catalogid)
                    + ";"
                )


class MWM_SNC_100pc_bright_BOSS_single_Carton(MWM_SNC_100pc_BOSS_Carton):
    """mwm_snc_100pc_bright_boss_single
    Shorthand name: mwm_snc_100pc_bright_boss_single

    Existing carton code:
    https://github.com/sdss/target_selection/blob/main/python/
    target_selection/cartons/mwm_snc.py Simplified Description of selection
    criteria :  Select everything within 100pc (i.e. parallax > 10mas) and
    Gaia-G fainter than 10th mag. Add some constraints on astrometric excess
    noise in crowded regions (Galactic plane, LMC, SMC, towards Galactic
    center). For details, see here. NOTE this is a subcarton of
    mwm_snc_100pc_boss (see above) where the only change is the cadence and
    priority. So, this should be able to inherit `class
    MWM_SNC_100pc_BOSS_Carton()`

    Return columns: Same columns as before + astrometric_excess_noise,
    phot_bp_rp_excess_factor, ruwe, phot_bp_mean_mag , phot_rp_mean_mag

    Metadata:
    Priority: 1807
    Cadence:  if G < 18, bright_1x1 ; if G > 18, dark_1x1
    Instrument: BOSS
    can_offset = True

    Lead contact: Ilija Medan
    """

    name = "mwm_snc_100pc_bright_boss_single"
    category = "science"
    program = "mwm_snc"
    mapper = "MWM"
    instrument = "BOSS"
    cadence = None  # cadence is set in post_process()
    priority = 1807
    can_offset = True

    def post_process(self, model):
        cursor = self.database.execute_sql(
            "select catalogid, phot_g_mean_mag from "
            + " sandbox.temp_mwm_snc_100pc_bright_boss_single "
            + " order by catalogid;"
        )

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_phot_g_mean_mag = output[i][1]

            if current_phot_g_mean_mag < 18:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_snc_100pc_bright_boss_single "
                    + " set cadence = 'bright_1x1' "
                    + " where catalogid = "
                    + str(current_catalogid)
                    + ";"
                )
            else:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_snc_100pc_bright_boss_single "
                    + " set cadence = 'dark_1x1' "
                    + " where catalogid = "
                    + str(current_catalogid)
                    + ";"
                )


class MWM_SNC_100pc_low_lum__BOSS_Carton(MWM_SNC_100pc_BOSS_Carton):
    """mwm_snc_100pc_low_lum_boss_single
    Shorthand name: mwm_snc_100pc_low_lum_boss_single

    Existing carton code:
    https://github.com/sdss/target_selection/blob/main/python/
    target_selection/cartons/mwm_snc.py Simplified Description of selection
    criteria :  Select everything within 100pc (i.e. parallax > 10mas) and
    Gaia-G fainter than 10th mag. Add some constraints on astrometric excess
    noise in crowded regions (Galactic plane, LMC, SMC, towards Galactic
    center). For details, see here. NOTE this is a subcarton of
    mwm_snc_100pc_boss (see above). So, this should be able to inherit
    `class MWM_SNC_100pc_BOSS_Carton()` and then add the below condition in
    red.

    Pseudo-code: The below text
    is what will need to be added on top of the BOSS base carton
    (`class MWM_SNC_100pc_BOSS_Carton()`)
     & (phot_g_mean_mag + 5 * log10(0.001 * parallax) + 5 > 16)

    Return columns: Same columns as before + astrometric_excess_noise,
    phot_bp_rp_excess_factor, ruwe, phot_bp_mean_mag , phot_rp_mean_mag

    Metadata:
    Priority: 1310
    Cadence:  dark_1x1
    Instrument: BOSS
    can_offset = True

    Lead contact: Ilija Medan
    """

    name = "mwm_snc_100pc_low_lum_boss_single"
    category = "science"
    program = "mwm_snc"
    mapper = "MWM"
    instrument = "BOSS"
    cadence = "dark_1x1"
    priority = 1310
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(
            Gaia_DR3.phot_g_mean_mag + 5 * peewee.fn.log10(0.001 * Gaia_DR3.parallax) + 5 > 16
        )

        return query

    # We have set the cadence above so we do not want to run
    # the post_process() of the base class MWM_SNC_100pc_BOSS_Carton.
    # Hence, we have the below do-nothing post_process().
    def post_process(self, model, **kwargs):
        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-02-16
# @Filename: too.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from sdssdb.peewee.sdss5db.catalogdb import CatalogToToO_Target, ToO_Metadata, ToO_Target
from sdssdb.peewee.sdss5db.targetdb import Carton, CartonToTarget, Target, Version

from .base import BaseCarton


__all__ = ["ToO_Carton"]


class ToO_Carton(BaseCarton):
    """Target of opportunity carton.

    Selects all the targets in ``catalogdb.too_target`` that don't yet exist in
    the carton.

    """

    name = "too"
    category = "science"
    cadence = "bright_1x1"
    priority = 3000
    program = "too"
    can_offset = True

    def build_query(self, version_id, query_region=None):
        C2TT = CatalogToToO_Target

        too_in_carton = (
            Target.select(Target.catalogid)
            .join(CartonToTarget)
            .join(Carton)
            .join(Version)
            .where(Version.plan == self.plan, Carton.carton == self.name)
        ).alias("too_in_carton")

        query = (
            ToO_Target.select(
                C2TT.catalogid,
                ToO_Target.fiber_type.alias("instrument"),
                ToO_Metadata.g_mag.alias("g"),
                ToO_Metadata.r_mag.alias("r"),
                ToO_Metadata.i_mag.alias("i"),
                ToO_Metadata.z_mag.alias("z"),
                ToO_Metadata.optical_prov,
            )
            .join(C2TT, on=(ToO_Target.too_id == C2TT.target_id))
            .switch(ToO_Target)
            .join(ToO_Metadata, on=(ToO_Target.too_id == ToO_Metadata.too_id))
            .where(
                C2TT.version_id == version_id,
                C2TT.best >> True,
                C2TT.catalogid.not_in(too_in_carton),
            )
        )

        return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Tom Dwelly
# @Date: 2020-03-31
# @Filename: bhm_rm.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
# derived from guide.py

# isort: skip_file

import peewee
from peewee import JOIN
from peewee import fn

from target_selection.cartons.base import BaseCarton
from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    BHM_RM_v1_3,
    CatalogToLegacy_Survey_DR8,
    CatalogToLegacy_Survey_DR10,
    CatalogToGaia_DR2,
    CatalogToGaia_DR3,
    CatalogToPanstarrs1,
)


#  This module provides the following BHM cartons in v1.0:
#  bhm_rm_core
#  bhm_rm_known_spec
#  bhm_rm_var
#  bhm_rm_ancillary
#  bhm_rm_xrayqso


class BhmRmBaseCarton(BaseCarton):
    """
    This class provides common setting and the masking routines used by all RM cartons
    """

    name = "bhm_rm_base"
    base_name = "bhm_rm_base"
    category = "science"
    mapper = "BHM"
    program = "bhm_rm"
    instrument = "BOSS"
    tile = False
    priority = None
    inertial = True
    alias_c = None
    alias_t = None
    alias_tw = None
    can_offset = False

    def get_fieldlist(self):
        """Read the RM field centres from the yaml"""
        fieldlist = []
        base_parameters = self.config["parameters"].get(self.base_name, None)
        if base_parameters:
            fieldlist = base_parameters["fieldlist"]
        return fieldlist

    def append_spatial_query(self, query, fieldlist):
        """extend the peewee query using a list of field centres"""
        if fieldlist is None:
            return query
        elif len(fieldlist) == 0:
            return query

        q = False
        for f in fieldlist:
            q = q | peewee.fn.q3c_radial_query(
                self.alias_c.ra, self.alias_c.dec, f["racen"], f["deccen"], f["radius"]
            )
        return query.where(q)

    def build_query(self, version_id, query_region=None):
        c = Catalog.alias()
        c2ls8 = CatalogToLegacy_Survey_DR8.alias()
        c2ls10 = CatalogToLegacy_Survey_DR10.alias()
        c2g3 = CatalogToGaia_DR3.alias()
        c2g2 = CatalogToGaia_DR2.alias()
        c2ps = CatalogToPanstarrs1.alias()
        t = BHM_RM_v1_3.alias()
        self.alias_c = c
        self.alias_t = t

        fieldlist = self.get_fieldlist()

        # fold in tiers of magnitude-based priority
        priority_mag_step = 0.5
        priority_mag_bright = 17.0
        priority_mag_faint = 22.0
        priority_mag_bright_known_spec = 20.5
        priority_floor = self.parameters.get("priority", 10000)
        priority1 = peewee.Case(
            None,
            (
                ((t.mag_i <= priority_mag_bright), priority_floor + 0),
                (
                    (
                        (self.name == "bhm_rm_known_spec")
                        & ~(t.rm_field_name.contains("SDSS-RM"))
                        & (t.mag_i <= priority_mag_bright_known_spec)
                    ),
                    priority_floor + 0,
                ),
                (
                    (t.mag_i <= priority_mag_faint),
                    priority_floor
                    + 5
                    * (
                        1
                        + peewee.fn.floor(
                            (t.mag_i - priority_mag_bright) / priority_mag_step
                        ).cast("int")
                    ),
                ),
                ((t.mag_i > priority_mag_faint), priority_floor + 95),
            ),
            None,
        )

        # combine the priorities
        priority = priority1

        value = peewee.Value(self.parameters.get("value", 1.0)).cast("float")
        instrument = peewee.Value(self.instrument)
        inertial = peewee.Value(self.inertial).cast("bool")

        # This is the scheme used in v0
        cadence_v0 = peewee.Case(
            None, ((t.rm_field_name.contains("S-CVZ"), "bhm_rm_lite5_100x8"),), "bhm_rm_174x8"
        )

        # this gives the new names for the same cadences assumed in v0
        cadence_v0p5 = peewee.Case(
            None, ((t.rm_field_name.contains("S-CVZ"), "dark_100x8"),), "dark_174x8"
        )

        # the following will replace old generic cadences when relevant table has been populated
        # TODO - replace when correct cadences are loaded
        cadence_v1p0 = peewee.Case(
            None,
            (
                (t.rm_field_name.contains("SDSS-RM"), "bhm_rm_sdss-rm"),
                (t.rm_field_name.contains("COSMOS"), "bhm_rm_cosmos"),
                (t.rm_field_name.contains("XMM-LSS"), "bhm_rm_xmm-lss"),
                (t.rm_field_name.contains("S-CVZ"), "bhm_rm_cvz-s"),
                (t.rm_field_name.contains("CDFS"), "bhm_rm_cdfs"),
                (t.rm_field_name.contains("ELIAS-S1"), "bhm_rm_elias-s1"),
            ),
            "dark_174x8",
        )

        opt_prov = peewee.Value("psfmag")

        route_taken = peewee.Case(
            None,
            (
                (c2ls10.catalogid.is_null(False), "lsdr10"),
                (c2g3.catalogid.is_null(False), "gdr3"),
                (c2ls8.catalogid.is_null(False), "lsdr8"),
                (c2g2.catalogid.is_null(False), "gdr2"),
                (c2ps.catalogid.is_null(False), "ps1dr2"),
            ),
            "unknown",
        )
        query = (
            t.select(
                c.catalogid,
                c2ls10.catalogid.alias("c2ls10_catalogid"),  # extra
                c2g3.catalogid.alias("c2g3_catalogid"),  # extra
                c2ls8.catalogid.alias("c2ls8_catalogid"),  # extra
                c2g2.catalogid.alias("c2g2_catalogid"),  # extra
                c2ps.catalogid.alias("c2ps_catalogid"),  # extra
                c.ra,  # extra
                c.dec,  # extra
                t.rm_field_name.alias("rm_field_name"),  # extra
                t.pkey.alias("rm_pkey"),  # extra
                instrument.alias("instrument"),
                priority.alias("priority"),
                value.alias("value"),
                cadence_v0p5.alias("cadence"),
                cadence_v0.alias("cadence_v0"),  # extra
                cadence_v0p5.alias("cadence_v0p5"),  # extra
                cadence_v1p0.alias("cadence_v1p0_maybe"),  # extra
                t.mag_g.alias("g"),
                t.mag_r.alias("r"),
                t.mag_i.alias("i"),
                t.mag_z.alias("z"),
                t.gaia_g.alias("gaia_g"),
                t.gaia_bp.alias("bp"),
                t.gaia_rp.alias("rp"),
                opt_prov.alias("optical_prov"),
                inertial.alias("inertial"),
                c2ls10.best.alias("c2ls10_best"),  # extra
                c2g3.best.alias("c2g3_best"),  # extra
                c2ls8.best.alias("c2ls8_best"),  # extra
                c2g2.best.alias("c2g2_best"),  # extra
                c2ps.best.alias("c2ps_best"),  # extra
                t.catalogidv05.alias("rm_catalogidv05"),  # extra
                t.ra.alias("rm_ra"),  # extra
                t.dec.alias("rm_dec"),  # extra
                route_taken.alias("route_taken"),  # extra
            )
            .join(c2ls10, JOIN.LEFT_OUTER, on=(c2ls10.target_id == t.ls_id_dr10))
            .join(c2g3, JOIN.LEFT_OUTER, on=(c2g3.target_id == t.gaia_dr3_source_id))
            .join(c2ls8, JOIN.LEFT_OUTER, on=(c2ls8.target_id == t.ls_id_dr8))
            .join(c2g2, JOIN.LEFT_OUTER, on=(c2g2.target_id == t.gaia_dr2_source_id))
            .join(c2ps, JOIN.LEFT_OUTER, on=(c2ps.target_id == t.panstarrs1_catid_objid))
            .join(
                c,
                on=(
                    fn.coalesce(
                        c2ls10.catalogid,
                        c2g3.catalogid,
                        c2ls8.catalogid,
                        c2g2.catalogid,
                        c2ps.catalogid,
                    )
                    == c.catalogid
                ),
            )
            .where(
                c.version_id == version_id,
                (
                    ((route_taken == "lsdr10") & (c2ls10.version_id == version_id))
                    | ((route_taken == "gdr3") & (c2g3.version_id == version_id))
                    | ((route_taken == "lsdr8") & (c2ls8.version_id == version_id))
                    | ((route_taken == "gdr2") & (c2g2.version_id == version_id))
                    | ((route_taken == "ps1dr2") & (c2ps.version_id == version_id))
                ),
                # the method below was throwing out ~20 cases where the lsdr8 ls_id
                # was unexpectedly unmatched to a catalogid in the v1 crossmatch
                # fn.coalesce(c2ls10.version_id, version_id) == version_id,
                # fn.coalesce(c2g3.version_id, version_id) == version_id,
                # fn.coalesce(c2ls8.version_id, version_id) == version_id,
                # fn.coalesce(c2g2.version_id, version_id) == version_id,
                # fn.coalesce(c2ps.version_id, version_id) == version_id,
                # ## fn.coalesce(c2ls10.best,True) >> True   # TODO check if this is dropping RM
                # ##                                         # targets like it does for AQMES
            )
            .where(
                (
                    (t.mag_i >= self.parameters["mag_i_min"])
                    & (t.mag_i < self.parameters["mag_i_max"])
                )
                | (
                    # S-CVZ targets often have only Gaia photom
                    (t.rm_field_name.contains("S-CVZ"))
                    & (t.gaia_g >= self.parameters["mag_g_min_cvz_s"])
                    & (t.gaia_g < self.parameters["mag_g_max_cvz_s"])
                )
            )
            .where(
                # Reject any objects where the t.rm_unsuitable flag is set
                t.rm_unsuitable >> False,
            )
            .distinct([t.pkey])  # avoid duplicates - trust the RM parent sample
            # - only needed if NOT using c2t.best = True condition
        )
        query = self.append_spatial_query(query, fieldlist)

        return query


class BhmRmCoreCarton(BhmRmBaseCarton):
    name = "bhm_rm_core"

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        t = self.alias_t
        query = query.where(
            (t.rm_core >> True),
            ~(t.rm_field_name.contains("SDSS-RM")),  # ignore this carton in the SDSS-RM field
        )

        return query


class BhmRmKnownSpecCarton(BhmRmBaseCarton):
    """
    bhm_rm_known_spec:  select all spectroscopically confirmed QSOs where redshift is extragalactic
    """

    name = "bhm_rm_known_spec"

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        t = self.alias_t
        query = query.where(
            ((t.rm_known_spec >> True),),
            (
                ~(t.rm_field_name.contains("SDSS-RM"))
                |
                # include extra constraints on SDSS-RM targets
                (t.mag_i < self.parameters["mag_i_max_sdss_rm"])
            ),
            (
                ~(t.rm_field_name.contains("COSMOS"))
                |
                # include extra constraints on COSMOS targets
                (t.mag_i < self.parameters["mag_i_max_cosmos"])
            ),
            (
                ~(t.rm_field_name.contains("XMM-LSS"))
                |
                # include extra constraints on XMM-LSS targets
                (t.mag_i < self.parameters["mag_i_max_xmm_lss"])
            ),
        )

        return query


class BhmRmVarCarton(BhmRmBaseCarton):
    """bhm_rm_var: selected based on g-band variability > 0.05 mag
    and bright enough to be detected by Gaia (G<~21)
    """

    name = "bhm_rm_var"

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        t = self.alias_t
        query = query.where(
            (t.rm_var >> True),
            ~(t.rm_field_name.contains("SDSS-RM")),  # ignore this carton in the SDSS-RM field
        )

        return query


class BhmRmAncillaryCarton(BhmRmBaseCarton):
    """
    bhm_rm_ancillary: from the Gaia_unWISE AGN catalog or the XDQSO catalog,
                      but requiring no proper motion/parallax detection from Gaia DR2
    """

    name = "bhm_rm_ancillary"

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        t = self.alias_t

        query = query.where(
            (t.rm_ancillary >> True),
            ~(t.rm_field_name.contains("SDSS-RM")),  # ignore this carton in the SDSS-RM field
        )

        return query


class BhmRmXrayQsoCarton(BhmRmBaseCarton):
    """
    bhm_rm_xrayqso:
       selected based on X-ray and SED

    """

    name = "bhm_rm_xrayqso"

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        t = self.alias_t
        query = query.where(
            (t.rm_xrayqso > 0),
            ~(t.rm_field_name.contains("SDSS-RM")),  # ignore this carton in the SDSS-RM field
        )

        return query



import glob
import importlib
import inspect
import os
import warnings

from astropy.utils.exceptions import AstropyDeprecationWarning

from ..exceptions import TargetSelectionImportWarning
from .base import BaseCarton


warnings.filterwarnings("ignore", category=AstropyDeprecationWarning)


# Import cartons so that they can be discovered by
# calling Carton.__subclasses__().

exclusions = ["__init__.py", "base.py"]

cwd = os.getcwd()
os.chdir(os.path.dirname(os.path.realpath(__file__)))

files = [file_ for file_ in glob.glob("*.py") if file_ not in exclusions]

for file_ in files:
    try:
        modname = file_[0:-3].replace("/", ".")
        mod = importlib.import_module("target_selection.cartons." + modname)
        for objname in dir(mod):
            obj = getattr(mod, objname)
            if inspect.isclass(obj) and issubclass(obj, BaseCarton):
                locals().update({objname: obj})

    except Exception as ee:
        warnings.warn(f"cannot import file {file_}: {ee}", TargetSelectionImportWarning)

os.chdir(cwd)
